import torch
import sys
import warnings
import argparse
import itertools
import triton
from aiter.ops.triton.attention.mha import (
    flash_attn_func,
    flash_attn_varlen_func,
    mha_set_use_fused_bwd_kernel,
)
from aiter.ops.triton.attention.mha_v3 import (
    flash_attn_fp8_func,
    flash_attn_varlen_fp8_func,
)
from aiter.test_mha_common import (
    generate_random_padding_mask,
    generate_qkv,
)
from op_tests.op_benchmarks.triton.utils.argparse import get_parser
from op_tests.op_benchmarks.triton.utils.benchmark_utils import (
    get_model_configs,
    print_vgpr,
    get_caller_name_no_ext,
)


def nonvarlen_benchmark_configs():
    batch_sizes = [1, 4, 16]
    N_HEADS = [16, 48]
    seq_len_q = [1, 1024, 4096]
    seq_len_k = [163, 8192]
    configs = list(itertools.product(batch_sizes, N_HEADS, seq_len_q, seq_len_k))
    configs = [
        (batch_size, N_HEAD, N_HEAD, seq_len_q, seq_len_k)
        for batch_size, N_HEAD, seq_len_q, seq_len_k in configs
    ]
    return configs


def varlen_benchmark_configs():
    batch_sizes = [1, 4, 8]
    N_HEADS = [16, 48]
    seq_len_q = [1, 1024, 4096]
    seq_len_k = [163, 8192]
    configs = list(itertools.product(batch_sizes, N_HEADS, seq_len_q, seq_len_k))
    configs = [
        (batch_size, N_HEAD, N_HEAD, seq_len_q, seq_len_k)
        for batch_size, N_HEAD, seq_len_q, seq_len_k in configs
    ]
    return configs


def model_benchmark_configs(args):
    config_file = args.model_configs
    configs = get_model_configs(config_path=config_file, models=args.model)
    fa_configs = []
    batch_size = args.b if args.b else 1

    for model_name, config in configs.items():
        HQ = config["num_attention_heads"]
        HK = (
            HQ
            if config["num_key_value_heads"] is None
            else config["num_key_value_heads"]
        )
        N_CTX_Q = args.sq if args.sq else [2**i for i in range(1, 14)]
        N_CTX_K = args.sk if args.sk else N_CTX_Q
        HEAD_DIM = config["hidden_size"] // HQ
        if isinstance(N_CTX_Q, list):
            for seq_len in N_CTX_Q:
                fa_configs.append(
                    (model_name, batch_size, HQ, HK, seq_len, seq_len, HEAD_DIM)
                )
        else:
            fa_configs.append(
                (model_name, batch_size, HQ, HK, N_CTX_Q, N_CTX_K, HEAD_DIM)
            )

    return fa_configs


def pad_rearrange_dropout_mask(
    S_dmask,
    cu_seqlens_q,
    cu_seqlens_k,
    max_seqlen_q,
    max_seqlen_k,
    seqlen_q,
    seqlen_k,
    num_q_heads,
):
    batch_size = cu_seqlens_q.numel() - 1

    padded_dropout_mask = torch.ones(
        (batch_size, num_q_heads, seqlen_q, seqlen_k), device="cuda"
    )
    for b in range(batch_size):
        start_q = cu_seqlens_q[b].item()
        end_q = cu_seqlens_q[b + 1].item()
        start_k = cu_seqlens_k[b].item()
        end_k = cu_seqlens_k[b + 1].item()

        seqlen_q = end_q - start_q
        seqlen_k = end_k - start_k
        for h in range(S_dmask.shape[1]):
            padded_dropout_mask[b, h, :max_seqlen_q, :max_seqlen_k] = S_dmask[
                b, h, :, :
            ]

    return padded_dropout_mask


def create_benchmark_configs(custom, args):
    dtype = arg_to_torch_dtype[args.dtype]
    hk = args.hq if not args.hk else args.hk
    sk = args.sq if not args.sk else args.sk
    head_size = 128 if not args.d else args.d
    head_size_v = head_size if not args.dv else args.dv
    mode = args.mode
    x_names = ["BATCH", "HQ", "HK", "N_CTX_Q", "N_CTX_K"]
    causal = args.causal
    varlen = args.layout == "thd"

    configs = []
    plot_name = get_caller_name_no_ext()
    extra_args = {
        "D_HEAD": head_size,
        "D_HEAD_V": head_size_v,
        "dtype": dtype,
        "causal": causal,
        "mode": mode,
    }

    if custom:
        x_vals_list = [(args.b, args.hq, hk, args.sq, sk)]
    else:
        if varlen:
            x_vals_list = varlen_benchmark_configs()  # Assume this exists
        else:
            x_vals_list = nonvarlen_benchmark_configs()  # Assume this exists

        if args.model:
            x_vals_list = model_benchmark_configs(args)
            x_names = ["model", "BATCH", "HQ", "HK", "N_CTX_Q", "N_CTX_K", "D_HEAD"]
            plot_name = f"fused-attention-{mode}-layout-{args.layout}-fp8-{args.fp8}-causal-{causal}"
            extra_args = {"dtype": dtype, "causal": causal, "mode": mode}

    if args.metric == "time":
        unit = "ms"
    elif args.metric == "throughput":
        unit = "TFLOPS"
    elif args.metric == "bandwidth":
        unit = "GB/s"
    else:
        raise ValueError("Unknown metric: " + args.metric)

    if mode == "bwd":
        if args.fused_bwd:
            line_vals = [f"fused-bwd({unit})"]
        else:
            line_vals = [f"onekernel-bwd({unit})"]
    else:
        line_vals = [f"fwd({unit})"]

    if args.bench_torch:
        line_vals = [f"Triton({unit})", f"Torch({unit})"]

    if args.test_mode:
        if args.fused_bwd:
            line_vals = [f"fused-bwd({unit})"]
        else:
            line_vals = [f"onekernel-bwd({unit})"]

    configs.append(
        triton.testing.Benchmark(
            x_names=x_names,
            x_vals=x_vals_list,
            line_arg="provider",
            line_vals=line_vals,
            line_names=line_vals,
            styles=[("red", "-"), ("green", "-"), ("yellow", "-")],
            ylabel=unit,
            plot_name=plot_name,
            args=extra_args,
        )
    )
    return configs


def run_benchmark(custom, args):
    torch.manual_seed(20)

    @triton.testing.perf_report(create_benchmark_configs(custom, args))
    def bench_mha(
        BATCH,
        HQ,
        HK,
        N_CTX_Q,
        N_CTX_K,
        D_HEAD,
        D_HEAD_V,
        dtype,
        causal,
        mode,
        provider,
        dropout=0.0,
        model=None,
        sm_scale=None,
        device="cuda",
    ):
        """
        Benchmark or test function for multi-head attention backward pass.
        In test_mode, verifies output matching with non-varlen inputs.
        """
        assert dropout <= 0.0, "Dropout not supported in this benchmark."
        requires_grad = mode == "bwd" or args.test_mode
        return_lse = True
        return_attn_probs = False
        varlen = args.layout == "thd"
        has_pe = D_HEAD > D_HEAD_V
        assert not (
            args.fp8 and has_pe
        ), "Positional Encoding (PE) doesn't support FP8 data type."
        assert not (
            has_pe and "fused-bwd" in provider
        ), "'Fused' backward implementation doesn't support Positional Encoding (PE)."
        assert not (
            args.fp8 and args.sink
        ), "Attention sink doesn't support FP8 data type."
        assert not (
            args.sink and "fused-bwd" in provider
        ), "'Fused' backward implementation doesn't support Attention Sink."

        global _USE_FUSED_BWD
        fused_backward = "fused-bwd" in provider
        mha_set_use_fused_bwd_kernel(fused_backward)

        # Default softmax scale to match standard attention
        if sm_scale is None:
            sm_scale = 1.0 / (D_HEAD**0.5)

        # Test mode: run tests from op_tests with specified shapes
        if args.test_mode:
            import op_tests.triton_tests.attention.test_mha as test_mha

            print(
                f"Testing kernel implementation <{provider}> against Torch with shape:"
            )
            print(
                f"BATCH={BATCH}, HQ={HQ}, HK={HK}, N_CTX_Q={N_CTX_Q}, N_CTX_K={N_CTX_K}, D_HEAD={D_HEAD}, D_HEAD_V={D_HEAD_V}"
            )
            if not varlen:
                if not has_pe:
                    test_mha.test_mha(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        dropout,
                        True,
                        True,
                        causal,
                        args.fp8,
                        dtype,
                    )
                else:
                    test_mha.test_mha_with_pe(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        D_HEAD_V,
                        dropout,
                        causal,
                    )
                print("Forward test passed!")
                if not has_pe:
                    test_mha.test_mha_backward_varlen(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        dropout,
                        causal,
                        args.fp8,
                        dtype,
                    )
                else:
                    test_mha.test_mha_backward_with_pe(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        D_HEAD_V,
                        dropout,
                        causal,
                    )
                print("Backward test passed!")
            else:
                if not has_pe:
                    test_mha.test_mha_varlen(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        dropout,
                        True,
                        True,
                        causal,
                        args.fp8,
                        dtype,
                    )
                else:
                    test_mha.test_mha_varlen_with_pe(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        D_HEAD_V,
                        dropout,
                        causal,
                    )
                print("Forward test passed!")
                if not has_pe:
                    test_mha.test_mha_backward(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        dropout,
                        causal,
                        args.fp8,
                        dtype,
                    )
                else:
                    test_mha.test_mha_backward_varlen_with_pe(
                        BATCH,
                        N_CTX_Q,
                        N_CTX_K,
                        HQ,
                        HK,
                        D_HEAD,
                        D_HEAD_V,
                        dropout,
                        causal,
                    )
                print("Backward test passed!")

            return 0

        # Generate base inputs
        q = torch.randn(
            (BATCH, N_CTX_Q, HQ, D_HEAD),
            device=device,
            dtype=dtype,
            requires_grad=requires_grad,
        )
        k = torch.randn(
            (BATCH, N_CTX_K, HK, D_HEAD),
            device=device,
            dtype=dtype,
            requires_grad=requires_grad,
        )
        v = torch.randn(
            (BATCH, N_CTX_K, HK, D_HEAD_V),
            device=device,
            dtype=dtype,
            requires_grad=requires_grad,
        )
        sink = (
            torch.randn((HQ,), device=device, dtype=dtype, requires_grad=requires_grad)
            if args.sink
            else None
        )

        # FLOPS calculation variables
        total_flops = 0.0

        # Input preparation
        if varlen:
            query_padding_mask = generate_random_padding_mask(
                N_CTX_Q, BATCH, device, mode="full" if args.equal_seqlens else "random"
            )
            key_padding_mask = generate_random_padding_mask(
                N_CTX_K, BATCH, device, mode="full" if args.equal_seqlens else "random"
            )
            (
                q_unpad,
                k_unpad,
                v_unpad,
                cu_seqlens_q,
                cu_seqlens_k,
                max_seqlen_q,
                max_seqlen_k,
                q,
                k,
                v,
                _,
                _,
                _,
            ) = generate_qkv(
                q, k, v, query_padding_mask, key_padding_mask, kvpacked=False
            )
            q_unpad.requires_grad = requires_grad
            k_unpad.requires_grad = requires_grad
            v_unpad.requires_grad = requires_grad

            q_input, k_input, v_input = q_unpad, k_unpad, v_unpad

            num_contexts = len(cu_seqlens_q) - 1
            for i in range(num_contexts):
                seqlen_q = (cu_seqlens_q[i + 1] - cu_seqlens_q[i]).item()
                seqlen_k = (cu_seqlens_k[i + 1] - cu_seqlens_k[i]).item()
                if causal:
                    valid_out_elements = (
                        ((seqlen_k**2 + seqlen_k) / 2)
                        if seqlen_q > seqlen_k
                        else (seqlen_q * seqlen_k - ((seqlen_q**2 - seqlen_q) / 2))
                    )
                    total_flops += valid_out_elements * HQ * (D_HEAD + D_HEAD_V) * 2.0
                else:
                    total_flops += seqlen_q * seqlen_k * HQ * (D_HEAD + D_HEAD_V) * 2.0
        else:
            q_input, k_input, v_input = q, k, v

            if causal:
                valid_out_elements = (
                    ((N_CTX_K**2 + N_CTX_K) / 2)
                    if N_CTX_Q > N_CTX_K
                    else (N_CTX_Q * N_CTX_K - ((N_CTX_Q**2 - N_CTX_Q) / 2))
                )
                total_flops += (
                    2.0 * BATCH * HQ * valid_out_elements * (D_HEAD + D_HEAD_V)
                )
            else:
                total_flops += (
                    2.0 * BATCH * HQ * N_CTX_Q * N_CTX_K * (D_HEAD + D_HEAD_V)
                )

        # Benchmark mode
        if varlen:
            if args.fp8:

                def fn():
                    return flash_attn_varlen_fp8_func(
                        q_input,
                        k_input,
                        v_input,
                        cu_seqlens_q,
                        cu_seqlens_k,
                        max_seqlen_q,
                        max_seqlen_k,
                        softmax_scale=sm_scale,
                        causal=causal,
                    )

            else:

                def fn():
                    return flash_attn_varlen_func(
                        q_input,
                        k_input,
                        v_input,
                        cu_seqlens_q,
                        cu_seqlens_k,
                        max_seqlen_q,
                        max_seqlen_k,
                        dropout_p=dropout,
                        softmax_scale=sm_scale,
                        causal=causal,
                        return_lse=return_lse,
                        return_attn_probs=return_attn_probs,
                        sink=sink,
                    )

        else:
            if args.fp8:

                def fn():
                    return flash_attn_fp8_func(
                        q_input,
                        k_input,
                        v_input,
                        softmax_scale=sm_scale,
                        causal=causal,
                    )

            else:

                def fn():
                    return flash_attn_func(
                        q_input,
                        k_input,
                        v_input,
                        dropout_p=dropout,
                        softmax_scale=sm_scale,
                        causal=causal,
                        return_lse=return_lse,
                        return_attn_probs=return_attn_probs,
                        sink=sink,
                    )

        if mode == "bwd":
            with torch.enable_grad():
                triton_out = fn()[0]
                d_out = torch.randn_like(triton_out)

                grad_inputs = (q_input, k_input, v_input)
                if sink is not None:
                    grad_inputs += (sink,)

                def fn():
                    grads = torch.autograd.grad(
                        triton_out,
                        grad_inputs,
                        d_out,
                        retain_graph=True,
                    )
                    return grads

        ms = triton.testing.do_bench(fn)

        if mode == "bwd":
            total_flops *= 2.5  # 2.0(bwd) + 0.5(recompute)

        if varlen:
            total_num_tokens_q = cu_seqlens_q[-1].item()
            total_num_tokens_k = cu_seqlens_k[-1].item()
        else:
            total_num_tokens_q = BATCH * N_CTX_Q
            total_num_tokens_k = BATCH * N_CTX_K
        q_size = total_num_tokens_q * HQ * D_HEAD * q.element_size()
        k_size = total_num_tokens_k * HK * D_HEAD * k.element_size()
        v_size = total_num_tokens_k * HK * D_HEAD_V * v.element_size()
        o_size = total_num_tokens_q * HQ * D_HEAD_V * q.element_size()
        if mode == "fwd":
            # read q, k, v
            mem_read = q_size + k_size + v_size
            # write o
            mem_write = o_size
        else:
            # read q, k, v, do
            mem_read = q_size + k_size + v_size + o_size
            # write dq, dk, dv
            mem_write = q_size + k_size + v_size
        mem = mem_read + mem_write

        # return ms
        if "ms" in provider:
            return ms
        elif "TFLOPS" in provider:
            return total_flops / ms * 1e-9
        else:  # GB/s
            return mem / ms * 1e-6

    bench_mha.run(save_path="." if args.o else None, print_data=True)


def supported_layouts():
    layouts = (
        "bshd: Q, K, V are individual tensors of [batch, seqlen_q/k, num_heads, head_size]. "
        "thd: Q, K, V are individual tensors of [total_q/k, num_heads, head_size]. "
    )
    return layouts


# argparse lacks support for boolean argument type (sigh...)
def str2bool(v):
    if isinstance(v, bool) or v is None:
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def parse_args():
    parser = get_parser(kernel_name="FlashAttention")
    parser.add_argument(
        "-mode", type=str, default="fwd", help="fwd:forward kernel, bwd:backward kernel"
    )
    parser.add_argument("-b", type=int, default=0)
    parser.add_argument("-hq", type=int, default=0)
    parser.add_argument("-hk", type=int, default=0)
    parser.add_argument("-sq", type=int, default=0)
    parser.add_argument("-sk", type=int, default=0)
    parser.add_argument(
        "-equal_seqlens",
        action="store_true",
        default=False,
        help="If specified, uses equal sequence lengths with thd layout, i.e t = b * sq",
    )
    parser.add_argument(
        "-d",
        type=int,
        default=0,
        help="Q and K head size, if -dv is absent then -d specifies V head size too",
    )
    parser.add_argument("-dv", type=int, default=0, help="optional V head size")
    parser.add_argument("-causal", type=str2bool, default=None)
    parser.add_argument("-fp8", action="store_true", default=False)
    parser.add_argument("-quantize_p", action="store_true", default=False)
    parser.add_argument("--dtype", default="fp16")
    parser.add_argument("-bench_torch", action="store_true", default=False)
    parser.add_argument("-fused_bwd", action="store_true", default=False)
    parser.add_argument("-print_vgpr", action="store_true", default=False)
    parser.add_argument(
        "-test_mode",
        action="store_true",
        default=False,
        help="Tests correctness of the Triton provider comparing the output to the Torch sdpa.",
    )

    parser.add_argument("--layout", type=str, default=None, help=supported_layouts())
    parser.add_argument(
        "-metric",
        nargs="?",
        const="throughput",
        choices=["time", "throughput", "bandwidth"],
        default=None,
        help="Metrics for the kernel benchmark.",
    )
    parser.add_argument(
        "-persistent",
        nargs="?",
        const="fixed",
        choices=["fixed", "dynamic"],
        default=None,
        help="Enable persistent kernels. Use '-persistent dynamic' for dynamic scheduling of the tiles.",
    )
    parser.add_argument(
        "-o", action="store_true", help="Write performance results to CSV file"
    )
    parser.add_argument(
        "-sink", action="store_true", default=False, help="use attention sink"
    )
    return parser.parse_args()


arg_to_torch_dtype = {
    "fp16": torch.float16,
    "bf16": torch.bfloat16,
    "fp32": torch.float32,
}


def main():
    args = parse_args()

    if args.model:
        if args.causal is None:  # User didn't specify -causal
            args.causal = True
        if args.layout is None:  # User didn't specify -layout
            args.layout = "thd"
        print(
            f"Note: using -model config defaults: causal={True}, layout={'thd'}. This is the most common real life scenario, but can be overridden with -causal and -layout flags."
        )
    else:
        # the defaults for causal and varlen when not using the -model
        if args.causal is None:  # User didn't specify -causal
            args.causal = False
        if args.layout is None:  # User didn't specify -layout
            args.layout = "bshd"

    custom_config = False

    assert (
        args.layout == "thd" or not args.equal_seqlens or args.model
    ), "Equal sequence lengths arg must be used with the thd layout or a model config."
    if args.hq or args.hk or args.d or args.dv:
        custom_config = True
        if not args.dv:
            args.dv = args.d
        assert (
            args.b and args.hq and args.sq and args.d and args.dv
        ), "If custom config is specified, please provide \
                all of batch, number of Q heads, Q sequence length \
                and head size."

    if args.model:
        assert not (
            args.hq or args.hk or args.d or args.dv
        ), "Specifying model fixes hq, hk and d already. Do not provide them!"

    assert (
        args.dtype in arg_to_torch_dtype
    ), "Only fp16, bf16 and f32 types currently supported."

    assert (
        args.layout in supported_layouts()
    ), f"{args.layout} is not in supported layouts: {supported_layouts()}."

    if args.layout == "thd" and args.equal_seqlens:
        warnings.warn(
            "Using 'thd' layout with equal_seqlen=True incurs an extra sequence length lookup cost "
            "compared to 'bshd' layout. Consider using 'bshd' for better performance.",
            category=RuntimeWarning,
        )

    if args.print_vgpr:
        assert not args.bench_torch, "Do not use -bench_torch with -print_vgpr."
        print("Retrieving VGPR usage for Triton kernels...")

        def fun():
            return run_benchmark(custom_config, args)

        print_vgpr(fun, get_caller_name_no_ext())
        return 0

    run_benchmark(custom_config, args)


if __name__ == "__main__":
    import sys

    sys.exit(main())



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.

import argparse
import logging
import os
from multiprocessing import Pool, freeze_support, set_start_method
from typing_extensions import Optional

import torch
import torch.distributed as dist

from aiter import dtypes
from aiter.dist.communication_op import tensor_model_parallel_all_reduce
from aiter.dist.parallel_state import (
    destroy_distributed_environment,
    destroy_model_parallel,
    ensure_model_parallel_initialized,
    get_tp_group,
    graph_capture,
    init_distributed_environment,
    set_custom_all_reduce,
)
from aiter.dist.utils import get_distributed_init_method, get_ip, get_open_port
from aiter.test_common import benchmark, checkAllclose, perftest

logger = logging.getLogger("aiter")

set_start_method("spawn", force=True)


def allreduce_custom(
    tp_size,
    pp_size,
    rankID,
    x,
    withGraph=False,
    distributed_init_method: Optional[str] = None,
):
    device = torch.device(f"cuda:{rankID}")
    torch.cuda.set_device(device)
    # init
    logger.info(f"RANK: {rankID} {tp_size} init_process_group...")
    set_custom_all_reduce(True)
    init_distributed_environment(
        world_size=tp_size,
        rank=rankID,
        distributed_init_method=distributed_init_method,
    )
    ensure_model_parallel_initialized(tp_size, pp_size)
    x = x.to(device)
    # dist.barrier(device_ids=[i for i in range(tp_size)])

    # warmup and align all gpu
    group = get_tp_group().device_group
    dist.all_reduce(torch.zeros(1).cuda(), group=group)
    torch.cuda.synchronize()

    if withGraph:
        graph = torch.cuda.CUDAGraph()
        with graph_capture() as gc:
            with torch.cuda.graph(graph, stream=gc.stream):
                out = tensor_model_parallel_all_reduce(x)
        out.fill_(0)

        @perftest()
        def run_ca():
            graph.replay()

        _, us = run_ca()
        out = (out, us)
    else:

        @perftest()
        def run_ca(x):
            return tensor_model_parallel_all_reduce(x)

        out = run_ca(x)

    # destroy
    if dist.is_initialized():
        destroy_model_parallel()
        destroy_distributed_environment()
        torch.cuda.empty_cache()
    return out


@benchmark()
def test_allreduce_custom(
    tp_size,
    pp_size,
    shape,
    dtype,
    withGraph=False,
    distributed_init_method: Optional[str] = None,
):
    os.environ["MASTER_ADDR"] = "127.0.0.1"
    os.environ["MASTER_PORT"] = "49373"
    pool = Pool(processes=tp_size)
    ref = torch.zeros(shape, dtype=dtype)
    rets = []
    for i in range(tp_size):
        x = torch.randn(shape, dtype=dtype)
        ref += x
        rets.append(
            pool.apply_async(
                allreduce_custom,
                args=(tp_size, pp_size, i, x, withGraph, distributed_init_method),
            )
        )
    pool.close()
    pool.join()
    rets = [el.get() for el in rets]
    for out, us in rets:
        msg = f"test_allreduce_custom: {shape=} {dtype=} {withGraph=} {us:>8.2f}"
        checkAllclose(ref, out.to(ref), msg=msg)


l_dtype = ["fp16", "bf16"]
l_shape = [(128, 8192)]

parser = argparse.ArgumentParser(description="config input of test")
parser.add_argument(
    "-d",
    "--dtype",
    type=str,
    choices=l_dtype,
    nargs="?",
    const=None,
    default=None,
    help="data type",
)
parser.add_argument(
    "-s",
    "--shape",
    type=dtypes.str2tuple,
    nargs="?",
    const=None,
    default=None,
    help="shape. e.g. -s 128,8192",
)
parser.add_argument(
    "-g",
    "--with-graph",
    type=lambda x: str(x).lower() in ["true", "1", "yes"],
    default=True,
    help="use CUDA graph (default: True). e.g. -g true or -g false",
)


if __name__ == "__main__":
    freeze_support()
    args = parser.parse_args()
    if args.dtype is None:
        l_dtype = [dtypes.d_dtypes[key] for key in l_dtype]
    else:
        l_dtype = [dtypes.d_dtypes[args.dtype]]
    if args.shape is not None:
        l_shape = [args.shape]
    for dtype in l_dtype:
        for shape in l_shape:
            test_allreduce_custom(
                8,
                1,
                shape,
                dtype,
                withGraph=args.with_graph,
                distributed_init_method=get_distributed_init_method(
                    get_ip(), get_open_port()
                ),
            )
            # test_allreduce_custom(8, 1, shape, dtype, withGraph=False)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.
# Adapted from flash-linear-attention: Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

"""
Chunk-based gated delta rule forward computation.

This module implements the chunk-based parallel computation for the gated delta rule.
Note: Only forward pass is implemented. Backward pass is not supported in aiter.
"""

import torch

from .chunk_delta_h import chunk_gated_delta_rule_fwd_h
from .chunk_o import chunk_fwd_o
from ..utils import (
    chunk_local_cumsum,
    chunk_scaled_dot_kkt_fwd,
    recompute_w_u_fwd,
    solve_tril,
)


def chunk_gated_delta_rule_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    g: torch.Tensor,
    beta: torch.Tensor,
    scale: float,
    initial_state: torch.Tensor,
    output_final_state: bool,
    cu_seqlens: torch.LongTensor | None = None,
):
    """
    Chunk gated delta rule forward computation (Forward only).

    This function implements chunk-based parallel computation for the gated delta rule,
    combining all necessary steps for efficient sequence processing.

    Note: This implementation only supports forward pass. Backward pass is not available.

    Args:
        q: Query tensor of shape [B, T, H, K]
        k: Key tensor of shape [B, T, H, K]
        v: Value tensor of shape [B, T, H, V]
        g: Gate tensor (in log space) of shape [B, T, H]
        beta: Beta parameter tensor of shape [B, T, H]
        scale: Scaling factor for queries
        initial_state: Initial hidden state of shape [N, H, K, V]
        output_final_state: Whether to output the final state
        cu_seqlens: Cumulative sequence lengths for variable-length inputs (optional) [N+1]

    Returns:
        tuple: (g, o, A, final_state) where:
            - g: Cumulative gate values [B, T, H]
            - o: Output tensor [B, T, H, V]
            - A: WY representation matrix
            - final_state: Final hidden state [N, H, K, V] if output_final_state=True, else None
    """
    # Step 1: Compute local cumulative sum of gates
    g = chunk_local_cumsum(g, chunk_size=64, cu_seqlens=cu_seqlens)

    # Step 2: Compute WY representation
    A = chunk_scaled_dot_kkt_fwd(
        k=k,
        g=g,
        beta=beta,
        cu_seqlens=cu_seqlens,
        output_dtype=torch.float32,
    )
    A = solve_tril(
        A=A,
        cu_seqlens=cu_seqlens,
        output_dtype=k.dtype,
    )
    w, u = recompute_w_u_fwd(
        k=k,
        v=v,
        beta=beta,
        A=A,
        g=g,
        cu_seqlens=cu_seqlens,
    )

    # Step 3: Compute hidden states
    h, v_new, final_state = chunk_gated_delta_rule_fwd_h(
        k=k,
        w=w,
        u=u,
        g=g,
        initial_state=initial_state,
        output_final_state=output_final_state,
        cu_seqlens=cu_seqlens,
    )

    # Step 4: Compute output
    o = chunk_fwd_o(
        q=q,
        k=k,
        v=v_new,
        h=h,
        g=g,
        scale=scale,
        cu_seqlens=cu_seqlens,
    )

    return g, o, A, final_state



import triton
import triton.language as tl
from aiter.ops.triton._triton_kernels.rope.rope import (
    _get_gptj_rotated_x_1D,
    _get_neox_rotated_x_1D,
)


@triton.jit
def _unit_cat(
    x1_ptr,
    x2_ptr,
    x_out_ptr,
    b,
    h,
    d1_offs,
    d2_offs,
    x1_stride_b,
    x1_stride_h,
    x1_stride_d,
    x2_stride_b,
    x2_stride_h,
    x2_stride_d,
    x_out_stride_b,
    x_out_stride_h,
    x_out_stride_d,
    BLOCK_D1: tl.constexpr,
):
    x1_offs = b * x1_stride_b + h * x1_stride_h + d1_offs * x1_stride_d
    x2_offs = b * x2_stride_b + h * x2_stride_h + d2_offs * x2_stride_d
    x_out_offs = b * x_out_stride_b + h * x_out_stride_h

    x1 = tl.load(x1_ptr + x1_offs)
    x2 = tl.load(x2_ptr + x2_offs)

    tl.store(x_out_ptr + x_out_offs + d1_offs * x_out_stride_d, x1)
    tl.store(x_out_ptr + x_out_offs + (d2_offs + BLOCK_D1) * x_out_stride_d, x2)


@triton.jit
def _qk_cat_kernel(
    q1_ptr,
    q2_ptr,
    k1_ptr,
    k2_ptr,
    q_out_ptr,
    k_out_ptr,
    q1_stride_b,
    q1_stride_h,
    q1_stride_d,
    q2_stride_b,
    q2_stride_h,
    q2_stride_d,
    k1_stride_b,
    k1_stride_h,
    k1_stride_d,
    k2_stride_b,
    k2_stride_h,
    k2_stride_d,
    q_out_stride_b,
    q_out_stride_h,
    q_out_stride_d,
    k_out_stride_b,
    k_out_stride_h,
    k_out_stride_d,
    QH_PER_KH: tl.constexpr,
    BLOCK_D1: tl.constexpr,
    BLOCK_D2: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_hq = tl.program_id(1)

    d1_offs = tl.arange(0, BLOCK_D1)
    d2_offs = tl.arange(0, BLOCK_D2)

    _unit_cat(
        q1_ptr,
        q2_ptr,
        q_out_ptr,
        pid_b,
        pid_hq,
        d1_offs,
        d2_offs,
        q1_stride_b,
        q1_stride_h,
        q1_stride_d,
        q2_stride_b,
        q2_stride_h,
        q2_stride_d,
        q_out_stride_b,
        q_out_stride_h,
        q_out_stride_d,
        BLOCK_D1,
    )

    if pid_hq % QH_PER_KH == 0:
        _unit_cat(
            k1_ptr,
            k2_ptr,
            k_out_ptr,
            pid_b,
            pid_hq // QH_PER_KH,
            d1_offs,
            d2_offs,
            k1_stride_b,
            k1_stride_h,
            k1_stride_d,
            k2_stride_b,
            k2_stride_h,
            k2_stride_d,
            k_out_stride_b,
            k_out_stride_h,
            k_out_stride_d,
            BLOCK_D1,
        )


@triton.jit
def _unit_rope_cat(
    x_nope_ptr,
    x_pe_ptr,
    cos,
    sin,
    x_out_ptr,
    b,
    h,
    d_nope_offs,
    d_pe_offs,
    x_nope_stride_b,
    x_nope_stride_h,
    x_nope_stride_d,
    x_pe_stride_b,
    x_pe_stride_h,
    x_pe_stride_d,
    x_out_stride_b,
    x_out_stride_h,
    x_out_stride_d,
    IS_NEOX: tl.constexpr,
    BLOCK_D_nope: tl.constexpr,
    BLOCK_D_pe: tl.constexpr,
    BLOCK_D_HALF_pe: tl.constexpr,
):
    x_nope_offs = (
        b * x_nope_stride_b + h * x_nope_stride_h + d_nope_offs * x_nope_stride_d
    )
    x_pe_offs = b * x_pe_stride_b + h * x_pe_stride_h + d_pe_offs * x_pe_stride_d
    x_out_offs = b * x_out_stride_b + h * x_out_stride_h

    x_nope = tl.load(x_nope_ptr + x_nope_offs)
    x_pe = tl.load(x_pe_ptr + x_pe_offs)

    if IS_NEOX:
        x_rotated_mask = d_pe_offs < BLOCK_D_HALF_pe
        x_pe_rotated = _get_neox_rotated_x_1D(
            x_pe, x_rotated_mask, BLOCK_D_pe, BLOCK_D_HALF_pe
        )
    else:
        x_rotated_mask = d_pe_offs % 2 == 0
        x_pe_rotated = _get_gptj_rotated_x_1D(
            x_pe, x_rotated_mask, BLOCK_D_pe, BLOCK_D_HALF_pe
        )

    x_pe = x_pe * cos + x_pe_rotated * sin
    x_pe = x_pe.to(x_pe_ptr.dtype.element_ty)

    tl.store(x_out_ptr + x_out_offs + d_nope_offs * x_out_stride_d, x_nope)
    tl.store(x_out_ptr + x_out_offs + (d_pe_offs + BLOCK_D_nope) * x_out_stride_d, x_pe)


@triton.jit
def _qk_rope_cat_kernel(
    q_nope_ptr,
    q_pe_ptr,
    k_nope_ptr,
    k_pe_ptr,
    pos_ptr,
    cos_ptr,
    sin_ptr,
    q_out_ptr,
    k_out_ptr,
    q_nope_stride_b,
    q_nope_stride_h,
    q_nope_stride_d,
    q_pe_stride_b,
    q_pe_stride_h,
    q_pe_stride_d,
    k_nope_stride_b,
    k_nope_stride_h,
    k_nope_stride_d,
    k_pe_stride_b,
    k_pe_stride_h,
    k_pe_stride_d,
    pos_stride_b,
    cos_stride_b,
    cos_stride_d,
    q_out_stride_b,
    q_out_stride_h,
    q_out_stride_d,
    k_out_stride_b,
    k_out_stride_h,
    k_out_stride_d,
    QH_PER_KH: tl.constexpr,
    REUSE_FREQS_FRONT_PART: tl.constexpr,
    IS_NEOX: tl.constexpr,
    BLOCK_D_nope: tl.constexpr,
    BLOCK_D_pe: tl.constexpr,
    BLOCK_D_HALF_pe: tl.constexpr,
):
    pid_b = tl.program_id(0)
    pid_hq = tl.program_id(1)

    d_nope_offs = tl.arange(0, BLOCK_D_nope)
    d_pe_offs = tl.arange(0, BLOCK_D_pe)

    if REUSE_FREQS_FRONT_PART:
        if IS_NEOX:
            d_cos_offs = d_pe_offs
            d_cos_offs = tl.where(
                (d_cos_offs >= BLOCK_D_HALF_pe) & (d_cos_offs < BLOCK_D_pe),
                d_cos_offs - BLOCK_D_HALF_pe,
                d_cos_offs,
            ).to(d_cos_offs.dtype)
            # d_cos_mask = d_cos_offs < BLOCK_D_pe
        else:
            d_cos_offs = d_pe_offs // 2
            # d_cos_mask = d_cos_offs < BLOCK_D_HALF_pe
    else:
        d_cos_offs = d_pe_offs
        # d_cos_mask = d_cos_offs < BLOCK_D_pe

    pos = tl.load(pos_ptr + pid_b * pos_stride_b)
    cos_offs = pos * cos_stride_b + d_cos_offs * cos_stride_d
    cos = tl.load(cos_ptr + cos_offs)
    sin = tl.load(sin_ptr + cos_offs)

    _unit_rope_cat(
        q_nope_ptr,
        q_pe_ptr,
        cos,
        sin,
        q_out_ptr,
        pid_b,
        pid_hq,
        d_nope_offs,
        d_pe_offs,
        q_nope_stride_b,
        q_nope_stride_h,
        q_nope_stride_d,
        q_pe_stride_b,
        q_pe_stride_h,
        q_pe_stride_d,
        q_out_stride_b,
        q_out_stride_h,
        q_out_stride_d,
        IS_NEOX,
        BLOCK_D_nope,
        BLOCK_D_pe,
        BLOCK_D_HALF_pe,
    )

    if pid_hq % QH_PER_KH == 0:
        _unit_rope_cat(
            k_nope_ptr,
            k_pe_ptr,
            cos,
            sin,
            k_out_ptr,
            pid_b,
            pid_hq // QH_PER_KH,
            d_nope_offs,
            d_pe_offs,
            k_nope_stride_b,
            k_nope_stride_h,
            k_nope_stride_d,
            k_pe_stride_b,
            k_pe_stride_h,
            k_pe_stride_d,
            k_out_stride_b,
            k_out_stride_h,
            k_out_stride_d,
            IS_NEOX,
            BLOCK_D_nope,
            BLOCK_D_pe,
            BLOCK_D_HALF_pe,
        )



from jinja2 import Template
from csrc.cpp_itfs.utils import (
    compile_template_op,
    transfer_hsaco,
    AITER_CORE_DIR,
    GPU_ARCH,
    get_default_func_name,
    not_built,
    run_lib,
)
from triton.tools.compile import compile_kernel, CompileArgs
import triton
import functools

MD_NAME = "asm_mla_decode_fwd"
warpSize = 64
with open(f"{AITER_CORE_DIR}/csrc/cpp_itfs/mla/asm_mla_decode_fwd.cpp.jinja", "r") as f:
    src_template = Template(f.read())

mgcs = {16: 64, 128: 16}


@functools.lru_cache()
def get_meta_param(num_kv_splits, device, bs):
    import torch

    if num_kv_splits is None:
        device_props = torch.cuda.get_device_properties(device)
        cu_num = device_props.multi_processor_count
        num_kv_splits = min(16, max(1, cu_num // bs))

    return num_kv_splits


def compile(
    gqa_ratio: int,
    page_size: int,
    q_dtype: str,
    kv_dtype: str,
    num_kv_splits: int,
    v_head_dim: int,
    func_name: str = None,
):
    if func_name is None:
        func_name = get_default_func_name(
            MD_NAME,
            (gqa_ratio, page_size, q_dtype, kv_dtype, num_kv_splits, v_head_dim),
        )

    if not_built(func_name):
        if gqa_ratio == 128:
            hsaco_path = f"{AITER_CORE_DIR}/hsa/{GPU_ARCH}/mla/mla_dec_stage1_bf16_a16w16_subQ128_mqa128.co"
            kernel_name = "_ZN5aiter41mla_dec_stage1_bf16_a16w16_subQ128_mqa128E"
        else:
            hsaco_path = f"{AITER_CORE_DIR}/hsa/{GPU_ARCH}/mla/mla_dec_stage1_bf16_a16w16_subQ16_mqa16.co"
            kernel_name = "_ZN5aiter39mla_dec_stage1_bf16_a16w16_subQ16_mqa16E"

        bin_size, bin_data = transfer_hsaco(hsaco_path)
        compile_args = CompileArgs(
            path=f"{AITER_CORE_DIR}/aiter/mla.py",
            kernel_name="_fwd_kernel_stage2_asm",
            signature=f"*fp32:16,*fp32:16,*bf16:16,*i32:16,*i32:16,i32,i32,i32,i32,i32,i32,i32,i32,{num_kv_splits},{triton.next_power_of_2(v_head_dim)},{v_head_dim},{mgcs[gqa_ratio]}",
            grid="bs,nheads,max_seqlen_q",
            num_warps=4,
            num_stages=2,
            out_name="decode_mla_stage2_asm",
        )
        triton_kernel, output_files = compile_kernel(compile_args)
        triton_header = None
        triton_source = None
        for output_file in output_files:
            if output_file.suffix == ".h":
                triton_header = output_file
            elif output_file.suffix == ".cpp":
                triton_source = output_file

        return compile_template_op(
            src_template,
            MD_NAME,
            ["../utils.h", "../../include", triton_header],
            [triton_source],
            bin_size=bin_size,
            bin_data=bin_data,
            page_size=page_size,
            q_dtype=q_dtype,
            kv_dtype=kv_dtype,
            triton_header=triton_header,
            kernel_name=kernel_name,
            triton_kernel=triton_kernel,
            num_kv_splits=num_kv_splits,
            func_name=func_name,
        )
    else:
        return run_lib(func_name)


def asm_mla_decode_fwd(
    q,  # [total_query_len, num_heads, head_size]
    kv_buffer,  # [num_page, page_size, num_kv_heads, kv_lora_rank + qk_rope_head_dim]
    output,  # [total_query_len, num_heads, head_size]
    qo_indptr,  # [num_seqs + 1]
    kv_indptr,  # [num_seqs + 1]
    kv_page_indices,  # [num_page_used]
    kv_last_page_lens,  # [batch_size]
    max_seqlen_q,
    softmax_scale=None,
    logit_cap=0.0,
    num_kv_splits=None,  # for experts only!!!
    logits=None,
    attn_lse=None,
):
    import torch
    from csrc.cpp_itfs.torch_utils import torch_to_c_types

    if q.dtype != torch.bfloat16:
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: only support dtype == torch.bfloat16 for now"
        )

    num_kv_heads = kv_buffer.size(2)
    total_query_len = output.size(0)
    num_heads = output.size(1)
    head_size = q.size(2)
    page_size = kv_buffer.size(1)
    v_head_dim = output.size(2)
    num_seqs = qo_indptr.size(0) - 1

    if num_kv_heads != 1:
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: only support num_kv_heads==1 for now"
        )

    if head_size != kv_buffer.size(3):
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: only support head_size == KV.size(3) for now"
        )

    if logit_cap > 0:
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: only support logit_cap==0 for now"
        )

    gqa_ratio = num_heads // num_kv_heads

    if softmax_scale is None:
        softmax_scale = 1.0 / (head_size**0.5)

    if num_kv_splits is None:
        num_kv_splits = get_meta_param(num_kv_splits, q.device, num_seqs)

    if logits is None:
        if num_heads == 16:
            logits = torch.empty(
                (total_query_len, num_kv_splits, num_heads, v_head_dim),
                dtype=torch.float,
                device=q.device,
            )
            if max_seqlen_q != 1:
                raise ValueError(
                    f"{asm_mla_decode_fwd.__name__}: only support max_seqlen_q==1 when n_head==16, but got {max_seqlen_q}"
                )
        elif num_heads == 128:
            logits = (
                output.view((total_query_len, num_kv_splits, num_heads, v_head_dim))
                if num_kv_splits == 1
                else torch.empty(
                    (total_query_len, num_kv_splits, num_heads, v_head_dim),
                    dtype=torch.float,
                    device=q.device,
                )
            )
        else:
            raise ValueError(
                f"{asm_mla_decode_fwd.__name__}: only support n_head==16 or n_head==128 for now"
            )
    if attn_lse is None:
        attn_lse = torch.empty(
            (total_query_len, num_kv_splits, num_heads, 1),
            dtype=torch.float,
            device=q.device,
        )

    if num_kv_splits != logits.size(1):
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: num_kv_splits != logits.size(1)"
        )

    if gqa_ratio == 16 and max_seqlen_q != 1:
        raise ValueError(
            f"{asm_mla_decode_fwd.__name__}: only support max_seqlen_q==1 when gqa_ratio==16"
        )

    func = compile(
        gqa_ratio,
        page_size,
        "__hip_bfloat16",
        "__hip_bfloat16",
        num_kv_splits,
        v_head_dim,
    )

    func(
        *torch_to_c_types(
            q,
            kv_buffer,
            qo_indptr,
            kv_indptr,
            kv_page_indices,
            kv_last_page_lens,
            max_seqlen_q,
            softmax_scale,
            logits,
            attn_lse,
            output,
            num_seqs,
            num_heads,
            num_kv_heads,
            q.stride(0),
            kv_buffer.stride(0),
            attn_lse.stride(0),
            attn_lse.stride(1),
            attn_lse.stride(2),
            output.stride(0),
            output.stride(1),
            torch.cuda.current_stream(q.device),
        )
    )
    if num_kv_splits == 1 and num_heads == 128:
        return logits.view(total_query_len, num_heads, v_head_dim), attn_lse
    return logits, attn_lse


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--gqa_ratio", type=int, required=True)
    parser.add_argument("--page_size", type=int, required=True)
    parser.add_argument("--q_dtype", type=str, required=True)
    parser.add_argument("--kv_dtype", type=str, required=True)
    parser.add_argument("--num_kv_splits", type=int, required=True)
    parser.add_argument("--v_head_dim", type=int, required=True)
    parser.add_argument("--func_name", type=str, default=None)
    args = parser.parse_args()
    compile(**vars(args))



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

import torch
import triton
import pytest
from aiter.ops.triton.gemm.fused.fused_gemm_afp4wfp4_a16w16 import (
    fused_gemm_afp4wfp4_a16w16,
)
from op_tests.triton_tests.gemm.basic.test_gemm_afp4wfp4 import (
    generate_gemm_afp4wfp4_inputs,
)
from op_tests.triton_tests.gemm.basic.test_gemm_afp4wfp4 import (
    run_torch as run_torch_fp4,
)
from op_tests.triton_tests.gemm.basic.test_gemm_a16w16 import (
    generate_gemm_a16w16_inputs,
)
import torch.nn.functional as F
import aiter.ops.triton.utils._triton.arch_info as arch_info


def run_torch(
    x_fp4,
    w_fp4,
    x_fp4_scale,
    w_fp4_scale,
    x_bf16,
    w_bf16,
    bias_fp4,
    bias_bf16,
    dtype=torch.bfloat16,
):
    y_fp4 = run_torch_fp4(x_fp4, w_fp4, x_fp4_scale, w_fp4_scale, dtype)
    if bias_fp4 is not None:
        y_fp4 += bias_fp4
    y_bf16 = F.linear(x_bf16, w_bf16, bias=bias_bf16)
    return y_fp4.to(dtype), y_bf16.to(dtype)


def run_triton(
    x_fp4,
    w_fp4,
    x_fp4_scale,
    w_fp4_scale,
    x_bf16,
    w_bf16,
    bias_fp4,
    bias_bf16,
    dtype=torch.bfloat16,
    y_fp4=None,
    y_bf16=None,
    skip_reduce=False,
    is_fp4_preshuffled=True,
):
    return fused_gemm_afp4wfp4_a16w16(
        x_fp4,
        w_fp4,
        x_fp4_scale,
        w_fp4_scale,
        x_bf16,
        w_bf16,
        is_fp4_preshuffled=is_fp4_preshuffled,
        bias_fp4=bias_fp4,
        bias_bf16=bias_bf16,
        dtype=dtype,
        y_fp4=y_fp4,
        y_bf16=y_bf16,
        skip_reduce=skip_reduce,
    )


def get_x_vals():

    x_vals = [
        (m, n1, n2, k)
        for k in [1024, 8192, 7168]
        for n2 in [256, 512]
        for n1 in [256, 512]
        for m in [1, 8, 32, 64, 128, 8192]
    ]
    return x_vals


@pytest.mark.parametrize("M, N1, N2, K", get_x_vals())
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
@pytest.mark.parametrize("output", [True, False])
@pytest.mark.parametrize("skip_reduce", [True, False])
@pytest.mark.parametrize("fp4_shuffle", [True, False])
def test_gemm(dtype, M, N1, N2, K, output, skip_reduce, fp4_shuffle):

    if not (arch_info.is_fp4_avail()):
        pytest.skip("MXFP4 not supported on this architecture")

    (
        x_fp4,
        w_fp4,
        w_fp4_triton,
        x_fp4_scale,
        w_fp4_scale,
        x_fp4_scale_triton,
        w_fp4_scale_triton,
        out_dtype,
        y_fp4,
    ) = generate_gemm_afp4wfp4_inputs(
        M,
        N1,
        K,
        dtype,
        layout="TN",
        output=output,
        shuffle_scales_fg=fp4_shuffle,
        shuffle_weight_fg=fp4_shuffle,
    )

    x_bf16, w_bf16, bias_bf16, _, y_bf16 = generate_gemm_a16w16_inputs(
        M, N2, K, dtype, output=output, bias=True
    )
    bias_bf16 = torch.randn((N2,), dtype=bias_bf16.dtype, device=bias_bf16.device)
    bias_fp4 = torch.randn((N1,), dtype=bias_bf16.dtype, device=bias_bf16.device)
    y_torch_fp4, y_torch_bf16 = run_torch(
        x_fp4,
        w_fp4,
        x_fp4_scale,
        w_fp4_scale,
        x_bf16,
        w_bf16,
        bias_fp4,
        bias_bf16,
        dtype,
    )
    y_triton_fp4, y_triton_bf16 = run_triton(
        x_fp4,
        w_fp4_triton,
        x_fp4_scale_triton,
        w_fp4_scale_triton,
        x_bf16,
        w_bf16,
        bias_fp4,
        bias_bf16,
        dtype,
        y_fp4,
        y_bf16,
        skip_reduce=skip_reduce,
        is_fp4_preshuffled=fp4_shuffle,
    )

    if y_triton_fp4.dim() == 3:
        y_triton_fp4 = y_triton_fp4.sum(axis=0).to(dtype=dtype)
        y_triton_bf16 = y_triton_bf16.sum(axis=0).to(dtype=dtype)

    triton.testing.assert_close(y_torch_bf16, y_triton_bf16, atol=0.1, rtol=1e-1)
    triton.testing.assert_close(y_torch_fp4, y_triton_fp4, atol=0.1, rtol=1e-1)



"""
Triton kernel helper functions shared across flash attention modules.

This module contains Triton JIT-compiled helper functions that are used within
the main attention kernels (fwd_prefill, fwd_decode, bwd). These are kept
separate from utils.py to allow stricter type checking on pure Python utilities.
"""

from typing import Literal, Optional, Tuple, Union

import torch
import triton
import triton.language as tl

from .utils import DEBUG, get_shape_from_layout, get_stride_from_layout, is_fp8


@triton.jit
def compute_alibi_block(
    alibi_slope, seqlen_q, seqlen_k, offs_m, offs_n, transpose=False
):
    """
    Compute ALiBi (Attention with Linear Biases) block.

    When seqlen_k and seqlen_q are different, the diagonal sticks to the
    bottom right of the attention matrix.
    """
    # e.g. alibi_slope = 1, seqlen_q = 2, seqlen_k = 5
    # offs_m = [0, 1], offs_n = [0, 1, 2, 3, 4]
    # Result: [[-3, -2, -1, 0, -1], [-4, -3, -2, -1, 0]]
    relative_pos_block = offs_m[:, None] + seqlen_k - seqlen_q - offs_n[None, :]
    alibi_block = -1 * alibi_slope * tl.abs(relative_pos_block)
    if transpose:
        return alibi_block.T
    else:
        return alibi_block


@triton.jit
def compute_fp8_scaling_factors(x, fp8_max: tl.constexpr):
    """Compute FP8 scaling and descaling factors for a block."""
    x_amax = tl.max(tl.abs(x))
    x_amax = tl.where(x_amax <= 1e-9, 1e-9, x_amax)
    scale_x = fp8_max / x_amax
    descale_x = x_amax / fp8_max
    return scale_x, descale_x


@triton.jit
def _cast_varlen_to_fp8_kernel_2d(
    X,
    X_fp8,
    Descale,
    cu_seqlens,
    H,
    MAX_SEQLEN,
    stride_batch,
    stride_seq,
    stride_head,
    stride_dim,
    stride_out_batch,
    stride_out_seq,
    stride_out_head,
    stride_out_dim,
    stride_desc_batch,
    stride_desc_head,
    FP8_CLAMP_VAL,
    FP8_MAX,
    BLOCK_SIZE: tl.constexpr,
    HEAD_DIM: tl.constexpr,
    ACTUAL_HEAD_DIM: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    """Cast tensor to FP8 with per-(batch, head) scaling."""
    b_id = tl.program_id(0)
    h_id = tl.program_id(1)

    # Get sequence bounds for this batch
    if IS_VARLEN:
        seq_start = tl.load(cu_seqlens + b_id)
        seq_end = tl.load(cu_seqlens + b_id + 1)
        seqlen = seq_end - seq_start
    else:
        seq_start = 0
        seqlen = MAX_SEQLEN

    # initialize max value tracker
    x_max_val = 0.0

    # STEP 1: Find max absolute value across the entire sequence
    num_of_blocks = tl.cdiv(seqlen, BLOCK_SIZE)
    for blk_idx in range(0, num_of_blocks):
        offs_seq = blk_idx * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
        offs_dim = tl.arange(0, HEAD_DIM)

        mask_seq = offs_seq[:, None] < seqlen
        if ACTUAL_HEAD_DIM != HEAD_DIM:
            mask_dim = offs_dim[None, :] < ACTUAL_HEAD_DIM
            mask_seq = mask_seq & mask_dim

        adj_x = (
            b_id * stride_batch
            + h_id * stride_head
            + seq_start * stride_seq
            + offs_seq[:, None] * stride_seq
            + offs_dim[None, :] * stride_dim
        )
        x_block = tl.load(X + adj_x, mask=mask_seq, other=0.0)
        block_max = tl.max(tl.abs(x_block))
        x_max_val = tl.maximum(x_max_val, block_max)

    # clamp to avoid division by zero
    x_max_val = tl.maximum(x_max_val, FP8_CLAMP_VAL)

    # compute scale and descale factors
    scale = FP8_MAX / x_max_val
    descale = x_max_val / FP8_MAX

    # store descale factor
    desc_ptr = Descale + b_id * stride_desc_batch + h_id
    tl.store(desc_ptr, descale)

    # STEP 2: Apply scaling and convert to FP8
    for blk_idx in range(0, num_of_blocks):
        offs_seq = blk_idx * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
        offs_dim = tl.arange(0, HEAD_DIM)

        mask_seq = offs_seq[:, None] < seqlen
        if ACTUAL_HEAD_DIM != HEAD_DIM:
            mask_dim = offs_dim[None, :] < ACTUAL_HEAD_DIM
            mask_seq = mask_seq & mask_dim

        addr = (
            b_id * stride_batch
            + h_id * stride_head
            + seq_start * stride_seq
            + offs_seq[:, None] * stride_seq
            + offs_dim[None, :] * stride_dim
        )
        x_block = tl.load(X + addr, mask=mask_seq, other=0.0)
        x_fp8_block = (x_block * scale).to(X_fp8.type.element_ty)

        addr_out = (
            b_id * stride_out_batch
            + h_id * stride_out_head
            + seq_start * stride_out_seq
            + offs_seq[:, None] * stride_out_seq
            + offs_dim[None, :] * stride_out_dim
        )
        tl.store(X_fp8 + addr_out, x_fp8_block, mask=mask_seq)


@triton.jit
def _rotary_kernel(
    OUT,
    X,
    COS,
    SIN,
    CU_SEQLENS,
    SEQLEN_OFFSETS,
    seqlen,
    nheads,
    seqlen_ro,
    stride_out_batch,
    stride_out_seqlen,
    stride_out_nheads,
    stride_out_headdim,
    stride_x_batch,
    stride_x_seqlen,
    stride_x_nheads,
    stride_x_headdim,
    ROTARY_DIM: tl.constexpr,
    IS_SEQLEN_OFFSETS_TENSOR: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    INTERLEAVED: tl.constexpr,
    CONJUGATE: tl.constexpr,
    BLOCK_H: tl.constexpr,
    BLOCK_M: tl.constexpr,
):
    """Apply rotary positional embeddings."""
    BLOCK_K: tl.constexpr = triton.next_power_of_2(ROTARY_DIM)
    ROTARY_DIM_HALF = ROTARY_DIM // 2
    pid_head = tl.program_id(axis=0)
    pid_m = tl.program_id(axis=1)
    pid_batch = tl.program_id(axis=2)

    if not IS_VARLEN:
        X = X + pid_batch * stride_x_batch
        OUT = OUT + pid_batch * stride_out_batch
    else:
        start_idx = tl.load(CU_SEQLENS + pid_batch)
        seqlen = tl.load(CU_SEQLENS + pid_batch + 1) - start_idx
        X = X + start_idx * stride_x_seqlen
        OUT = OUT + start_idx * stride_out_seqlen

    if pid_m * BLOCK_M >= seqlen:
        return

    rh = pid_head * BLOCK_H + tl.arange(0, BLOCK_H)
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    if not IS_SEQLEN_OFFSETS_TENSOR:
        rm_cs = rm + SEQLEN_OFFSETS
    else:
        rm_cs = rm + tl.load(SEQLEN_OFFSETS + pid_batch)

    rk_half = tl.arange(0, BLOCK_K // 2)
    COS = COS + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    SIN = SIN + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    mask_cs = (rm_cs[:, None] < seqlen_ro) & (rk_half[None, :] < ROTARY_DIM_HALF)
    cos = tl.load(COS, mask=mask_cs, other=1.0).to(tl.float32)
    sin = tl.load(SIN, mask=mask_cs, other=0.0).to(tl.float32)
    if CONJUGATE:
        sin = -sin

    if not INTERLEAVED:
        X = X + (
            rh[:, None, None] * stride_x_nheads
            + rm[None, :, None] * stride_x_seqlen
            + rk_half[None, None, :] * stride_x_headdim
        )
        OUT = OUT + (
            rh[:, None, None] * stride_out_nheads
            + rm[None, :, None] * stride_out_seqlen
            + rk_half[None, None, :] * stride_out_headdim
        )
        mask = (
            (rh[:, None, None] < nheads)
            & (rm[None, :, None] < seqlen)
            & (rk_half[None, None, :] < ROTARY_DIM_HALF)
        )
        x0 = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x1 = tl.load(X + ROTARY_DIM_HALF * stride_x_headdim, mask=mask, other=0.0).to(
            tl.float32
        )
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        tl.store(OUT, o0, mask=mask)
        tl.store(OUT + ROTARY_DIM_HALF * stride_out_headdim, o1, mask=mask)
    else:
        rk = tl.arange(0, BLOCK_K)
        X = X + (
            rh[:, None, None] * stride_x_nheads
            + rm[None, :, None] * stride_x_seqlen
            + rk[None, None, :] * stride_x_headdim
        )
        OUT = OUT + (
            rh[:, None, None] * stride_out_nheads
            + rm[None, :, None] * stride_out_seqlen
            + rk[None, None, :] * stride_out_headdim
        )
        mask = (
            (rh[:, None, None] < nheads)
            & (rm[None, :, None] < seqlen)
            & (rk[None, None, :] < ROTARY_DIM)
        )
        x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x0, x1 = tl.split(tl.reshape(x, [BLOCK_H, BLOCK_M, BLOCK_K // 2, 2]))
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        o = tl.reshape(tl.join(o0, o1), [BLOCK_H, BLOCK_M, BLOCK_K])
        tl.store(OUT, o, mask=mask)


# -------------------------------
# Python wrappers for Triton kernels
# -------------------------------


def cast_to_fp8(
    x: torch.Tensor,
    fp8_dtype: torch.dtype,
    layout: Literal["bshd", "thd"],
    clamp_val: float = 1e-9,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Cast tensor to FP8 with per-(batch, head) scaling factors."""
    if DEBUG > 0:
        print()
        print("cast_to_fp8")
        print("x:", x, x.shape)
        print("fp8_dtype:", fp8_dtype)
        print("cu_seqlens:", cu_seqlens)
        print("max_seqlen:", max_seqlen)
        print("clamp_val:", clamp_val)

    assert x.dtype in {
        torch.float16,
        torch.float32,
        torch.float64,
        torch.bfloat16,
    } and is_fp8(fp8_dtype), f"Cannot cast {x.dtype} to {fp8_dtype}"

    batch, max_seqlen_final, num_heads, head_dim = get_shape_from_layout(
        x, layout, cu_seqlens, max_seqlen
    )
    is_varlen = layout == "thd"
    fp8_max = torch.finfo(fp8_dtype).max

    padded_head_dim = 1 << (head_dim - 1).bit_length()
    padded_head_dim = max(padded_head_dim, 32)

    x_fp8 = torch.zeros_like(x, dtype=fp8_dtype)
    descale_factors = torch.zeros(
        (batch, num_heads), device=x.device, dtype=torch.float32
    )
    BLOCK_SIZE = 128

    stride_batch, stride_head, stride_seq, stride_dim = get_stride_from_layout(
        x, layout
    )
    stride_out_batch, stride_out_head, stride_out_seq, stride_out_dim = (
        get_stride_from_layout(x_fp8, layout)
    )
    stride_desc_batch, stride_desc_head = descale_factors.stride()

    grid = (batch, num_heads)
    _cast_varlen_to_fp8_kernel_2d[grid](
        x,
        x_fp8,
        descale_factors,
        cu_seqlens,
        num_heads,
        max_seqlen_final,
        stride_batch,
        stride_seq,
        stride_head,
        stride_dim,
        stride_out_batch,
        stride_out_seq,
        stride_out_head,
        stride_out_dim,
        stride_desc_batch,
        stride_desc_head,
        clamp_val,
        fp8_max,
        BLOCK_SIZE=BLOCK_SIZE,
        HEAD_DIM=padded_head_dim,
        ACTUAL_HEAD_DIM=head_dim,
        IS_VARLEN=is_varlen,
    )

    return x_fp8, descale_factors


def _apply_rotary_kernel(
    x: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
    interleaved: bool = False,
    inplace: bool = False,
    conjugate: bool = False,
) -> torch.Tensor:
    """Apply rotary positional embeddings using Triton kernel."""
    is_varlen = cu_seqlens is not None
    if not is_varlen:
        batch, seqlen, nheads, headdim = x.shape
    else:
        assert (
            max_seqlen is not None
        ), "If cu_seqlens is passed, max_seqlen must also be provided"
        total_seqlen, nheads, headdim = x.shape
        assert cu_seqlens is not None
        batch_p_1 = cu_seqlens.shape[0]
        batch = batch_p_1 - 1
        seqlen = max_seqlen
    seqlen_ro, rotary_dim_half = cos.shape
    assert sin.shape == cos.shape
    rotary_dim = 2 * rotary_dim_half
    assert rotary_dim <= headdim
    assert headdim <= 256
    assert seqlen_ro >= seqlen

    cos, sin = cos.contiguous(), sin.contiguous()
    if isinstance(seqlen_offsets, torch.Tensor):
        assert seqlen_offsets.shape == (batch,)
        assert seqlen_offsets.dtype in (torch.int32, torch.int64)
        seqlen_offsets = seqlen_offsets.contiguous()
    else:
        assert seqlen_offsets + seqlen <= seqlen_ro

    out = torch.empty_like(x) if not inplace else x
    if rotary_dim < headdim and not inplace:
        out[..., rotary_dim:].copy_(x[..., rotary_dim:])

    BLOCK_M = 8 if rotary_dim <= 128 else 4
    grid = (
        triton.cdiv(nheads, 2),
        triton.cdiv(seqlen, BLOCK_M),
        batch,
    )

    with torch.cuda.device(x.device.index):
        torch.library.wrap_triton(_rotary_kernel)[grid](
            out,
            x,
            cos,
            sin,
            cu_seqlens,
            seqlen_offsets,
            seqlen,
            nheads,
            seqlen_ro,
            out.stride(0) if not is_varlen else 0,
            out.stride(-3),
            out.stride(-2),
            out.stride(-1),
            x.stride(0) if not is_varlen else 0,
            x.stride(-3),
            x.stride(-2),
            x.stride(-1),
            rotary_dim,
            isinstance(seqlen_offsets, torch.Tensor),
            is_varlen,
            interleaved,
            conjugate,
            BLOCK_M=BLOCK_M,
            BLOCK_H=2,
        )
    return out


class _ApplyRotary(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        x: torch.Tensor,
        cos: torch.Tensor,
        sin: torch.Tensor,
        interleaved: bool,
        inplace: bool,
        seqlen_offsets: Union[int, torch.Tensor],
        cu_seqlens: Optional[torch.Tensor],
        max_seqlen: Optional[int],
    ) -> torch.Tensor:
        out = _apply_rotary_kernel(
            x,
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            cu_seqlens=cu_seqlens,
            max_seqlen=max_seqlen,
            interleaved=interleaved,
            inplace=inplace,
            conjugate=False,
        )
        if isinstance(seqlen_offsets, int):
            ctx.save_for_backward(cos, sin, cu_seqlens)
            ctx.seqlen_offsets = seqlen_offsets
        else:
            ctx.save_for_backward(cos, sin, cu_seqlens, seqlen_offsets)
            ctx.seqlen_offsets = None
        ctx.interleaved = interleaved
        ctx.inplace = inplace
        ctx.max_seqlen = max_seqlen
        return out if not inplace else x

    @staticmethod
    def backward(
        ctx, do: torch.Tensor
    ) -> tuple[torch.Tensor, None, None, None, None, None, None, None]:
        seqlen_offsets = ctx.seqlen_offsets
        if seqlen_offsets is None:
            cos, sin, cu_seqlens, seqlen_offsets = ctx.saved_tensors
        else:
            cos, sin, cu_seqlens = ctx.saved_tensors
        dx = _apply_rotary_kernel(
            do,
            cos,
            sin,
            seqlen_offsets=seqlen_offsets,
            cu_seqlens=cu_seqlens,
            max_seqlen=ctx.max_seqlen,
            interleaved=ctx.interleaved,
            inplace=ctx.inplace,
            conjugate=True,
        )
        return dx, None, None, None, None, None, None, None


def apply_rotary_emb(
    x: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    interleaved: bool = False,
    inplace: bool = False,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
) -> torch.Tensor:
    """Apply rotary embeddings to tensor x.

    Args:
        x: (B, S, H, D) if `cu_seqlens` is None else (total_S, H, D).
        cos, sin: (S_rotary, rotary_dim/2)
        interleaved: GPT-J style if True.
        inplace: modify x in place.
        seqlen_offsets: int or (B,) tensor of starting offsets per sequence.
        cu_seqlens: (B+1,) tensor enabling varlen mode.
        max_seqlen: required when `cu_seqlens` is provided.
    """
    original_dtype = x.dtype
    is_fp8_input = is_fp8(original_dtype)
    if is_fp8_input:
        target_dtype = (
            torch.bfloat16
            if cos.dtype == torch.bfloat16 or torch.cuda.is_bf16_supported()
            else torch.float16
        )
        x_up = x.to(target_dtype)
        cos_up = cos.to(target_dtype) if cos.dtype != target_dtype else cos
        sin_up = sin.to(target_dtype) if sin.dtype != target_dtype else sin
        out_up = _ApplyRotary.apply(
            x_up,
            cos_up,
            sin_up,
            interleaved,
            False,
            seqlen_offsets,
            cu_seqlens,
            max_seqlen,
        )
        if inplace:
            x.copy_(out_up.to(original_dtype))
            return x
        return out_up.to(original_dtype)
    else:
        return _ApplyRotary.apply(
            x, cos, sin, interleaved, inplace, seqlen_offsets, cu_seqlens, max_seqlen
        )


def apply_rotary(
    q: torch.Tensor,
    k_new: Optional[torch.Tensor],
    cos: torch.Tensor,
    sin: torch.Tensor,
    *,
    causal: bool,
    local: bool,
    interleaved: bool = False,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """Apply rotary embeddings to q and optionally k_new.

    Policy:
      - If causal OR local attention: apply rotary directly on (B, S, H, D).
      - Else (non-causal global): flatten heads into sequence, apply, unflatten.
      - k_new is always rotated directly when provided.
    """
    assert q.ndim == 4, f"Expected q shape (B,S,H,D), got {q.shape}"
    B, S, H, D = q.shape
    use_flatten = (not causal) and (not local)

    if use_flatten:
        q_flat = q.reshape(B, S * H, D).unsqueeze(1)
        q_flat = apply_rotary_emb(
            q_flat, cos, sin, interleaved=interleaved, seqlen_offsets=seqlen_offsets
        )
        q = q_flat.view(B, 1, S * H, D).reshape(B, S, H, D)
    else:
        q = apply_rotary_emb(
            q, cos, sin, interleaved=interleaved, seqlen_offsets=seqlen_offsets
        )

    if k_new is not None:
        k_new = apply_rotary_emb(
            k_new, cos, sin, interleaved=interleaved, seqlen_offsets=seqlen_offsets
        )
    return q, k_new



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.

import torch
import pytest
from aiter.ops.triton.gemm.fused.fused_gemm_a8w8_blockscale_split_cat import (
    fused_gemm_a8w8_blockscale_split_cat,
    fused_gemm_a8w8_blockscale_preshuffle_split_cat,
)

from aiter.ops.triton.utils.types import str_to_torch_dtype, get_fp8_dtypes
import torch.nn.functional as F

from aiter.ops.shuffle import shuffle_weight
import aiter.ops.triton.utils._triton.arch_info as arch_info

block_shape = (128, 128)
DEVICE_ARCH = arch_info.get_arch()


def run_torch(x, w, y, x_scale, w_scale, S1, S2, D, dtype=torch.bfloat16):
    block_shape_n, block_shape_k = block_shape
    m, c1 = x.shape
    n = w.shape[0]

    x_scale = x_scale.repeat_interleave(block_shape_k, dim=1)
    x = x.to(x_scale.dtype) * x_scale[:m, :c1]
    x = x.view(m, c1)

    w_scale = w_scale.repeat_interleave(block_shape_n, dim=0)
    w_scale = w_scale.repeat_interleave(block_shape_k, dim=1)
    w_scale = w_scale[:n, :c1]
    w = w.to(w_scale.dtype) * w_scale

    c = F.linear(x.to(torch.float32), w.to(torch.float32))
    c = c.view(-1, D, S1 + S2)
    c1, c2 = c.split([S1, S2], dim=-1)
    c1 = torch.cat([c1, y.expand((*c1.shape[:-1], -1))], dim=-1)

    return c1.to(dtype), c2.to(dtype)


def run_triton(impl, x, w, y, x_scale, w_scale, S1, S2, D, dtype=torch.bfloat16):
    m = x.shape[0]

    return impl(x, w, y.expand(m, D, -1), x_scale, w_scale, S1, S2, dtype)


e5m2_type, e4m3_type = get_fp8_dtypes()


def get_shapes():
    x_vals = [(1024 * v, 1024 * v, 1024 * v) for v in range(1, 9)]
    x_vals += [(4864, 4096, 8192), (9728, 8192, 65536)]
    x_vals += [
        (1, 1280, 8192),
        (32, 1280, 8192),
        (64, 1280, 8192),
        (128, 1280, 8192),
        (192, 1280, 8192),
        (256, 1280, 8192),
        (320, 1280, 8192),
        (512, 1280, 8192),
        (1024, 1280, 8192),
        (2048, 1280, 8192),
        (4096, 1280, 8192),
        (8192, 1280, 8192),
        (16384, 1280, 8192),
        (1, 8192, 1024),
        (32, 8192, 1024),
        (64, 8192, 1024),
        (128, 8192, 1024),
        (192, 8192, 1024),
        (256, 8192, 1024),
        (320, 8192, 1024),
        (512, 8192, 1024),
        (1024, 8192, 1024),
        (2048, 8192, 1024),
        (4096, 8192, 1024),
        (8192, 8192, 1024),
        (16384, 8192, 1024),
        (2048, 2048, 2049),
        (159, 17389, 597),
        (16, 576, 7168),
    ]
    x_vals += [
        (256, 8192, 1024),
        (256, 1024, 8192),
        (256, 32768, 8192),
        (256, 8192, 32768),
    ]
    x_vals = [
        (4096, 256, 512),
    ]
    return x_vals


def generate_fused_gemm_a8w8_blockscale_split_cat_inputs(
    M: int,
    N: int,
    K: int,
    S3: int,
    block_shape_n: int,
    block_shape_k: int,
    dtype=torch.bfloat16,
    layout: str = "TN",
    shuffle: bool = False,
):
    """
    The GEMM kernel expects:
    - x: (M, K) -> row-major format
    - w: (N, K) -> column-major format
    - y: (M, D, S3)
    """
    scale_n = (N + block_shape_n - 1) // block_shape_n
    scale_k = (K + block_shape_k - 1) // block_shape_k

    if layout[0] == "T":
        x = (torch.rand((M, K), dtype=torch.float16, device="cuda") / 10).to(e4m3_type)
    else:
        x = (
            (torch.rand((K, M), dtype=torch.float16, device="cuda") / 10)
            .to(e4m3_type)
            .T
        )

    if layout[1] == "N":
        w = (torch.rand((N, K), dtype=torch.float16, device="cuda") / 10).to(e4m3_type)
    else:
        w = (
            (torch.rand((K, N), dtype=torch.float16, device="cuda") / 10)
            .to(e4m3_type)
            .T
        )

    x_scale = torch.rand([M, scale_k], dtype=torch.float32, device="cuda")
    w_scale = torch.rand([scale_n, scale_k], dtype=torch.float32, device="cuda")

    if shuffle:
        w_shuffle_layout = (16, 16)
        w_shuffled = shuffle_weight(w, w_shuffle_layout).reshape(
            w.shape[0] // w_shuffle_layout[0],
            w.shape[1] * w_shuffle_layout[0],
        )
        x_scale_shuffled = x_scale.transpose(0, 1).contiguous().view(*x_scale.shape)
    else:
        w_shuffled = w
        x_scale_shuffled = x_scale

    y = torch.rand((M, S3), dtype=torch.bfloat16, device="cuda").unsqueeze(1)

    return x, w, w_shuffled, y, x_scale, x_scale_shuffled, w_scale


# TODO d, s3, layout = 32 128 TT UT failed if AMDGCN_USE_BUFFER_OPS=1 on MI300
@pytest.mark.parametrize(
    "dtype, M, N, K, D, S3, layout, impl",
    [
        (dtype, *shape, d, s3, layout, impl)
        for dtype in ["bf16"]
        for shape in get_shapes()
        for d in [16]
        for s3 in [64]
        for layout in ["TN", "TT", "NN", "NT"]
        for impl in ["triton", "triton_shuffle"]
    ],
)
def test_fused_gemm_a8w8_blockscale_split_cat(dtype, M, N, K, D, S3, layout, impl):
    torch.cuda.empty_cache()  # Helps avoid hangs in large tests
    torch.cuda.synchronize()

    block_shape_n, block_shape_k = block_shape

    # skip tests
    if K % block_shape_k != 0:
        pytest.skip(
            "Latest upstream compiler as of Aug 22 (necessary for Gluon) causes"
            " infinite hang when EVEN_K is false. Try seeing if it's fixed if it's been a while."
        )
    if N % D != 0:
        pytest.skip("N must be divisible by D as N = D * (S1 + S2)")
    if impl == "triton_shuffle":
        if N % 16 > 0 or K % 32 > 0:
            pytest.skip(
                "N has to be multiple of 16 and K has to be multiple of 32 for preshuffle cases"
            )

    # deconstruct N
    S = N // D
    S1 = S // 2
    S2 = S - S1

    dtype = str_to_torch_dtype[dtype]
    x, w, w_triton, y, x_scale, x_scale_triton, w_scale = (
        generate_fused_gemm_a8w8_blockscale_split_cat_inputs(
            M,
            N,
            K,
            S3,
            block_shape_n,
            block_shape_k,
            dtype=dtype,
            layout=layout,
            shuffle=("_shuffle" in impl),
        )
    )

    if impl == "triton":
        impl = fused_gemm_a8w8_blockscale_split_cat
    elif impl == "triton_shuffle":
        impl = fused_gemm_a8w8_blockscale_preshuffle_split_cat
    else:
        raise ValueError(f"Unknown implementation: {impl}")

    c1_torch, c2_torch = run_torch(x, w, y, x_scale, w_scale, S1, S2, D, dtype)
    c1_triton, c2_triton = run_triton(
        impl, x, w_triton, y, x_scale_triton, w_scale, S1, S2, D, dtype
    )

    torch.testing.assert_close(c1_torch, c1_triton, atol=0.01, rtol=1e-2)
    torch.testing.assert_close(c2_torch, c2_triton, atol=0.01, rtol=1e-2)



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.

import sys
import argparse
import random
from typing import Optional, Tuple, Union, Dict
import subprocess
import multiprocessing
import concurrent.futures
from multiprocessing import cpu_count

import pandas as pd
import torch
import triton
import triton.language as tl

import aiter
from aiter import dtypes
from aiter.test_common import benchmark
from csrc.cpp_itfs.utils import (
    BUILD_DIR,
    get_default_func_name,
)
from csrc.cpp_itfs.pa_gluon_aot.pa_decode_gluon_aot import (
    pa_decode_gluon_aot,
)

MD_NAME = "pa_decode_attention_reduce_kernel"

torch.set_default_device("cuda")
torch.set_printoptions(sci_mode=False)

# Triton to PyTorch dtype mapping
TL_TO_TORCH_DTYPE = {
    tl.float8e4b8: torch.float8_e4m3fnuz,
    tl.float8e4nv: torch.float8_e4m3fn,
    tl.bfloat16: torch.bfloat16,
    tl.float16: torch.float16,
}
TORCH_TO_TL_DTYPE = {
    torch.float8_e4m3fnuz: tl.float8e4b8,
    torch.float8_e4m3fn: tl.float8e4nv,
    torch.bfloat16: tl.bfloat16,
    torch.float16: tl.float16,
}

# Test configuration parameters
USE_TORCH_FLASH_REF_OPTIONS = [True]
USE_AOT_IMPL_OPTIONS = [True, False]
KV_VARLEN_OPTIONS = [False, True]
TRANS_V_OPTIONS = [False, True]
# QUANT_Q_AND_KV_OPTIONS = [[True, True], [False, False]]
QUANT_Q_AND_KV_OPTIONS = [[False, False]]
CONTEXT_PARTITION_SIZE_OPTIONS = [256]
COMPUTE_TYPE_OPTIONS = ["fp8", "bf16", "fp16"]
QUANT_MODE_OPTIONS = ["per_token", "per_tensor"]
HEAD_DIMENSION_OPTIONS = [128]

BLOCK_SIZE_OPTIONS = [16, 64, 1024]
HEAD_CONFIGURATIONS = [(5, 1), (8, 1), (10, 1), (16, 1), (64, 4)]
QUERY_LENGTH_OPTIONS = [1, 2, 3, 4]
CONTEXT_LENGTH_OPTIONS = [512, 4096, 4097]
BATCH_SIZE_OPTIONS = [4, 80, 128]
SINKS_OPTIONS = [True, False]
SLIDING_WINDOW_OPTIONS = [0, 128]
COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS = []


def run_gluon_kernel(
    output: torch.Tensor,
    query: torch.Tensor,
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    context_lengths: torch.Tensor,
    block_tables: torch.Tensor,
    softmax_scale: float,
    query_length: int,
    max_context_partition_num: int,
    context_partition_size: int,
    compute_type: torch.dtype,
    query_scale: torch.Tensor,
    key_scale: torch.Tensor,
    value_scale: torch.Tensor,
    exp_sums: torch.Tensor,
    max_logits: torch.Tensor,
    temporary_output: torch.Tensor,
    alibi_slopes: Optional[torch.Tensor] = None,
    sinks: Optional[torch.Tensor] = None,
    use_aot_impl: bool = False,
) -> None:
    """Run Gluon FP8/BF16/FP16 kernel for paged attention.

    Args:
        output: Output tensor [batch_size * query_length, num_query_heads, head_size]
        query: Query tensor [batch_size * query_length, num_query_heads, head_size]
        key_cache: Key cache tensor [num_blocks, num_kv_heads, head_size // x, kv_block_size, x]
        value_cache: Value cache tensor [num_blocks, num_kv_heads, head_size, kv_block_size] or [num_blocks, num_kv_heads, kv_block_size // x, head_size, x]
        context_lengths: Current context lengths for each sequence [num_seqs]
        block_tables: Mapping from sequences to physical cache blocks [num_seqs, max_num_blocks_per_seq]
        softmax_scale: Softmax scale (typically 1/sqrt(head_size))
        query_length: Query sequence length
        max_context_partition_num: Maximum number of context partitions
        context_partition_size: Context partition size
        compute_type: Compute data type (torch.dtype)
        query_scale: Query quantization scale
        key_scale: Key quantization scale
        value_scale: Value quantization scale
        exp_sums: Buffer for exponential sums
        max_logits: Buffer for maximum logits
        temporary_output: Buffer for partial outputs
        alibi_slopes: ALiBi slopes (optional)
        sinks: Sink tokens (optional)
        use_aot_impl: Whether to use AOT implementation

    Returns:
        None (modifies output in-place)

    This function can run in aot or jit mode based on use_aot_impl flag.
    """
    if use_aot_impl:
        pa_decode_gluon_aot(
            output,
            query,
            key_cache,
            value_cache,
            context_lengths,
            block_tables,
            softmax_scale,
            query_length,
            max_context_partition_num,
            context_partition_size,
            compute_type,
            query_scale,
            key_scale,
            value_scale,
            exp_sums=exp_sums,
            max_logits=max_logits,
            temporary_output=temporary_output,
            alibi_slopes=alibi_slopes,
            run_compiled_kernel=True,
            sinks=sinks,
        )


@benchmark()
def run_pa_gluon_test(
    context_length: int,
    batch_size: int,
    num_heads: Tuple[int, int],
    head_size: int,
    block_size: int,
    compute_type: torch.dtype,
    query_length: int,
    quant_mode: str,
    context_partition_size: int,
    trans_v: bool,
    kv_varlen: bool,
    use_aot_impl: bool,
    quant_q: bool,
    quant_kv: bool,
) -> Dict[str, Union[float, str]]:
    """Test paged attention decode with gluon implementations."""
    results = {}
    data_type = compute_type
    if compute_type == aiter.dtypes.fp8:
        data_type = torch.bfloat16
    device = "cuda:0"
    torch.set_default_device(device)
    num_query_heads, num_kv_heads = num_heads
    assert (
        num_query_heads % num_kv_heads == 0
    ), "Query heads must be divisible by KV heads"

    max_context_length = max(1024, context_length)
    max_blocks_per_sequence = (max_context_length + block_size - 1) // block_size
    total_blocks = max_blocks_per_sequence * batch_size
    blocks_per_sequence = (context_length + block_size - 1) // block_size
    total_queries = batch_size * query_length

    if kv_varlen:
        random.seed(123)
        kv_len_list = [
            random.randint(query_length, context_length) for _ in range(batch_size)
        ]
    else:
        kv_len_list = [context_length] * batch_size
    context_lengths = torch.tensor(kv_len_list, dtype=torch.int32, device=device)

    random.seed(123)
    block_tables_list = []
    for _ in range(batch_size):
        block_table = [
            random.randint(0, total_blocks - 1) for _ in range(blocks_per_sequence)
        ]
        block_tables_list.append(block_table)

    block_tables = torch.tensor(block_tables_list, dtype=torch.int32, device=device)

    # Create KV cache tensors directly with torch.empty
    elements_per_vector = 16 // data_type.itemsize
    key_cache_shape = (
        total_blocks,
        num_kv_heads,
        head_size // elements_per_vector,
        block_size,
        elements_per_vector,
    )
    key_cache = torch.empty(key_cache_shape, dtype=data_type, device=device)

    value_cache_shape = (total_blocks, num_kv_heads, head_size, block_size)
    value_cache = torch.empty(value_cache_shape, dtype=data_type, device=device)

    softmax_scale = 1.0 / (head_size**0.5)

    # Create query tensors directly
    query = torch.empty(
        total_queries, num_query_heads, head_size, dtype=data_type, device=device
    )

    # Quantization based on mode and flags
    if quant_mode == "per_token":
        # Per-token quantization for query (if enabled)
        if quant_q:
            quantized_query = torch.empty(
                total_queries,
                num_query_heads,
                head_size,
                dtype=aiter.dtypes.fp8,
                device=device,
            )
            # Per-token query scale factors: [total_queries, num_query_heads, 1]
            query_scale_factors = torch.empty(
                total_queries, num_query_heads, 1, dtype=torch.float32, device=device
            )
        else:
            quantized_query = query
            query_scale_factors = None

        # Per-token quantization for KV cache (if enabled)
        if quant_kv:
            quantized_keys = torch.empty(
                key_cache_shape, dtype=aiter.dtypes.fp8, device=device
            )
            # Determine value cache shape based on trans_v
            if trans_v:
                # Shuffled layout: [num_blocks, num_kv_heads, block_size // x, head_size, x]
                quant_elements_per_vector = 16 // aiter.dtypes.fp8.itemsize
                quantized_values_shape = (
                    total_blocks,
                    num_kv_heads,
                    block_size // quant_elements_per_vector,
                    head_size,
                    quant_elements_per_vector,
                )
            else:
                # Normal layout: [num_blocks, num_kv_heads, head_size, block_size]
                quantized_values_shape = value_cache_shape
            quantized_values = torch.empty(
                quantized_values_shape, dtype=aiter.dtypes.fp8, device=device
            )
            # Per-token KV scale factors: [num_blocks, num_kv_heads, block_size, 1]
            key_scale_original = torch.empty(
                total_blocks,
                num_kv_heads,
                block_size,
                1,
                dtype=torch.float32,
                device=device,
            )
            value_scale_original = torch.empty(
                total_blocks,
                num_kv_heads,
                block_size,
                1,
                dtype=torch.float32,
                device=device,
            )
        else:
            quantized_keys = key_cache
            # Determine value cache shape based on trans_v
            if trans_v:
                # Shuffled layout: [num_blocks, num_kv_heads, block_size // x, head_size, x]
                cache_elements_per_vector = 16 // data_type.itemsize
                quantized_values_shape = (
                    total_blocks,
                    num_kv_heads,
                    block_size // cache_elements_per_vector,
                    head_size,
                    cache_elements_per_vector,
                )
                quantized_values = torch.empty(
                    quantized_values_shape, dtype=data_type, device=device
                )
            else:
                quantized_values = value_cache
            key_scale_original = None
            value_scale_original = None
    else:  # per_tensor
        # Per-tensor quantization for query (if enabled)
        if quant_q:
            quantized_query = torch.empty(
                total_queries,
                num_query_heads,
                head_size,
                dtype=aiter.dtypes.fp8,
                device=device,
            )
            # Per-tensor query scale factor: scalar [1]
            query_scale_factors = torch.empty(1, dtype=torch.float32, device=device)
        else:
            quantized_query = query
            query_scale_factors = None

        # Per-tensor quantization for KV cache (if enabled)
        if quant_kv:
            quantized_keys = torch.empty(
                key_cache_shape, dtype=aiter.dtypes.fp8, device=device
            )
            # Determine value cache shape based on trans_v
            if trans_v:
                # Shuffled layout: [num_blocks, num_kv_heads, block_size // x, head_size, x]
                quant_elements_per_vector = 16 // aiter.dtypes.fp8.itemsize
                quantized_values_shape = (
                    total_blocks,
                    num_kv_heads,
                    block_size // quant_elements_per_vector,
                    head_size,
                    quant_elements_per_vector,
                )
            else:
                # Normal layout: [num_blocks, num_kv_heads, head_size, block_size]
                quantized_values_shape = value_cache_shape
            quantized_values = torch.empty(
                quantized_values_shape, dtype=aiter.dtypes.fp8, device=device
            )
            # Per-tensor KV scale factors: scalar [1]
            key_scale_original = torch.empty(1, dtype=torch.float32, device=device)
            value_scale_original = torch.empty(1, dtype=torch.float32, device=device)
        else:
            quantized_keys = key_cache
            # Determine value cache shape based on trans_v
            if trans_v:
                # Shuffled layout: [num_blocks, num_kv_heads, block_size // x, head_size, x]
                cache_elements_per_vector = 16 // data_type.itemsize
                quantized_values_shape = (
                    total_blocks,
                    num_kv_heads,
                    block_size // cache_elements_per_vector,
                    head_size,
                    cache_elements_per_vector,
                )
                quantized_values = torch.empty(
                    quantized_values_shape, dtype=data_type, device=device
                )
            else:
                quantized_values = value_cache
            key_scale_original = None
            value_scale_original = None

    # Test Gluon
    query_group_size = num_query_heads // num_kv_heads
    num_seqs = batch_size
    max_context_partition_num = (
        context_lengths.max().item() + context_partition_size - 1
    ) // context_partition_size
    query_group_size = num_query_heads // num_kv_heads
    equivalent_query_group_size = query_length * query_group_size
    intermediate_shape = (
        num_seqs,
        num_kv_heads,
        max_context_partition_num,
        equivalent_query_group_size,
    )
    exp_sums = torch.empty(intermediate_shape, dtype=torch.float32, device=device)
    max_logits = torch.empty(intermediate_shape, dtype=torch.float32, device=device)
    temporary_output = torch.empty(
        *intermediate_shape,
        head_size,
        dtype=data_type,
        device=device,
    )
    # Create output tensor for final results
    output = torch.empty(
        total_queries, num_query_heads, head_size, dtype=data_type, device=device
    )

    run_gluon_kernel(
        output,
        quantized_query,
        quantized_keys,
        quantized_values,
        context_lengths,
        block_tables,
        softmax_scale,
        query_length,
        max_context_partition_num,
        context_partition_size,
        compute_type,
        query_scale=query_scale_factors,
        key_scale=key_scale_original,
        value_scale=value_scale_original,
        exp_sums=exp_sums,
        max_logits=max_logits,
        temporary_output=temporary_output,
        alibi_slopes=None,
        sinks=None,
        use_aot_impl=use_aot_impl,
    )

    return results


def create_argument_parser() -> argparse.ArgumentParser:
    """Create command line argument parser."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Test paged attention decode gluon implementation",
    )

    parser.add_argument(
        "--compute_type",
        type=str,
        default=None,
        help="Compute type",
    )
    parser.add_argument(
        "-n",
        "--num_heads",
        type=dtypes.str2tuple,
        default=None,
        help="Number of heads (q_heads, kv_heads)",
    )
    parser.add_argument(
        "-q",
        "--query_length",
        type=int,
        choices=QUERY_LENGTH_OPTIONS,
        default=None,
        help="Query length",
    )
    parser.add_argument(
        "-c", "--context_length", type=int, default=None, help="Context length"
    )
    parser.add_argument("-b", "--batch_size", type=int, default=None, help="Batch size")
    parser.add_argument("--block_size", type=int, default=None, help="Block size")
    parser.add_argument(
        "--quant_mode",
        type=str,
        choices=["per_token", "per_tensor", "both"],
        default=None,
        help="Quantization mode: per_token, per_tensor, or both",
    )
    parser.add_argument(
        "--quant_q_and_kv",
        type=dtypes.str2tuple,
        default=None,
        help=(
            "Tuple of bools specifying whether to quant_q and quant_kv, e.g. 0,0 "
            "First value is for quant_q, second for quant_kv."
        ),
    )
    parser.add_argument(
        "--trans_v",
        type=lambda x: (str(x).lower() == "true"),
        default=None,
        help="Transpose value cache layout (True/False)",
    )
    parser.add_argument(
        "--kv_varlen",
        type=lambda x: (str(x).lower() == "true"),
        default=None,
        help="KV use varlen (True/False)",
    )
    parser.add_argument(
        "--use_torch_flash_ref",
        type=lambda x: (str(x).lower() == "true"),
        default=None,
        help="Use torch flash reference implementation (True/False)",
    )
    parser.add_argument(
        "--use_aot_impl",
        type=lambda x: (str(x).lower() == "true"),
        default=None,
        help="Use gluon AOT implementation (True/False)",
    )
    parser.add_argument(
        "--context_partition_size",
        type=int,
        default=None,
        help="Sequence partition size",
    )
    parser.add_argument(
        "--num_processes",
        type=int,
        default=32,
        help="Number of parallel processes to use (default: use 1 CPU cores)",
    )

    return parser


def process_arguments(args: argparse.Namespace) -> tuple:
    """Process command line arguments."""
    compute_types = COMPUTE_TYPE_OPTIONS
    block_sizes = BLOCK_SIZE_OPTIONS
    head_configs = HEAD_CONFIGURATIONS
    context_length = CONTEXT_LENGTH_OPTIONS
    batch_sizes = BATCH_SIZE_OPTIONS
    query_lengths = QUERY_LENGTH_OPTIONS
    quant_mode = QUANT_MODE_OPTIONS
    trans_v = TRANS_V_OPTIONS
    kv_varlen = KV_VARLEN_OPTIONS
    quant_q_and_kv = QUANT_Q_AND_KV_OPTIONS
    use_torch_flash_ref_options = USE_TORCH_FLASH_REF_OPTIONS
    use_aot_impl_options = USE_AOT_IMPL_OPTIONS
    context_partition_size_options = CONTEXT_PARTITION_SIZE_OPTIONS
    sinks_options = SINKS_OPTIONS
    sliding_window_options = SLIDING_WINDOW_OPTIONS

    if args.compute_type is not None:
        compute_types = [dtypes.d_dtypes[args.compute_type]]
    else:
        compute_types = [dtypes.d_dtypes[key] for key in compute_types]

    if args.num_heads is not None:
        head_configs = [args.num_heads]
    if args.query_length is not None:
        query_lengths = [args.query_length]
    if args.context_length is not None:
        context_length = [args.context_length]
    if args.batch_size is not None:
        batch_sizes = [args.batch_size]
    if args.block_size is not None:
        block_sizes = [args.block_size]
    if args.block_size is not None:
        block_sizes = [args.block_size]
    if args.quant_mode is not None:
        quant_mode = [args.quant_mode]
    if args.trans_v is not None:
        trans_v = [args.trans_v]
    if args.kv_varlen is not None:
        kv_varlen = [args.kv_varlen]
    if args.quant_q_and_kv is not None:
        quant_q_and_kv = [args.quant_q_and_kv]

    # Process new arguments
    if args.use_torch_flash_ref is not None:
        use_torch_flash_ref_options = [args.use_torch_flash_ref]
    if args.use_aot_impl is not None:
        use_aot_impl_options = [args.use_aot_impl]
    if args.context_partition_size is not None:
        context_partition_size_options = [args.context_partition_size]

    # Create compute_types_quant_q_and_kv list (same as test_pa_decode_gluon.py)
    compute_types_quant_q_and_kv = []
    for ct in compute_types:
        for quant_q, quant_kv in quant_q_and_kv:
            compute_types_quant_q_and_kv.append([ct, quant_q, quant_kv])
    if len(COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS) > 0:
        compute_types_quant_q_and_kv = COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS
        for idx in range(len(compute_types_quant_q_and_kv)):
            if not isinstance(compute_types_quant_q_and_kv[idx][0], torch.dtype):
                compute_types_quant_q_and_kv[idx][0] = dtypes.d_dtypes[
                    compute_types_quant_q_and_kv[idx][0]
                ]

    return (
        block_sizes,
        head_configs,
        context_length,
        batch_sizes,
        query_lengths,
        quant_mode,
        trans_v,
        kv_varlen,
        compute_types_quant_q_and_kv,
        use_torch_flash_ref_options,
        use_aot_impl_options,
        context_partition_size_options,
        sinks_options,
        sliding_window_options,
    )


def _run_single_test(args):
    """
    Helper function to run a single test case.

    Args:
        args: Tuple containing (test_config, current, total)

    Returns:
        Dictionary containing test results
    """
    test_config, current, total = args

    print(
        f"\n[{current}/{total}] Testing: "
        f"use_torch_flash_ref={test_config['use_torch_flash_ref']}, "
        f"compute_type={test_config['compute_type']}, "
        f"quant_q_and_kv=({test_config['quant_q']}, {test_config['quant_kv']}), "
        f"use_aot_impl={test_config['use_aot_impl']}, "
        f"trans_v={test_config['trans_v']}, "
        f"kv_varlen={test_config['kv_varlen']}, "
        f"context_partition_size={test_config['context_partition_size']}, "
        f"quant_mode={test_config['quant_mode']}, "
        f"block_size={test_config['block_size']}, "
        f"num_heads={test_config['num_heads']}, "
        f"context_lengths={test_config['context_length']}, "
        f"batch_size={test_config['batch_size']}, "
        f"query_length={test_config['query_length']}, "
        f"head_size={test_config['head_size']}"
    )

    result = run_pa_gluon_test(
        context_length=test_config["context_length"],
        batch_size=test_config["batch_size"],
        num_heads=test_config["num_heads"],
        head_size=test_config["head_size"],
        block_size=test_config["block_size"],
        compute_type=test_config["compute_type"],
        query_length=test_config["query_length"],
        quant_mode=test_config["quant_mode"],
        context_partition_size=test_config["context_partition_size"],
        trans_v=test_config["trans_v"],
        kv_varlen=test_config["kv_varlen"],
        use_aot_impl=test_config["use_aot_impl"],
        quant_q=test_config["quant_q"],
        quant_kv=test_config["quant_kv"],
    )

    return result


def run_multi_pa_gluon_test(
    block_sizes,
    head_configs,
    context_length,
    batch_sizes,
    query_lengths,
    quant_mode,
    trans_v,
    kv_varlen,
    compute_types_quant_q_and_kv,
    use_torch_flash_ref_options,
    use_aot_impl_options,
    context_partition_size_options,
    num_processes=None,
    sinks_options=[False],
    sliding_window_options=[0],
) -> pd.DataFrame:
    """
    Run all tests using multiprocessing for parallel execution.

    Args:
        num_processes: Number of parallel processes to use.
                      If None, uses cpu_count().

    Returns:
        DataFrame containing all test results
    """
    cdna_version = 3
    # Generate all test configurations
    test_configs = []

    for use_torch_flash_ref in use_torch_flash_ref_options:
        for hc in head_configs:
            for ct, quant_q, quant_kv in compute_types_quant_q_and_kv:
                for trans_v_mode in trans_v:
                    for kv_varlen_mode in kv_varlen:
                        for context_partition_size in context_partition_size_options:
                            qm_cnt = 0
                            for qm in quant_mode:
                                qm_cnt += 1
                                if not quant_q and not quant_kv and qm_cnt > 1:
                                    continue
                                for bs in block_sizes:
                                    for head_size in HEAD_DIMENSION_OPTIONS:
                                        for ql in query_lengths:
                                            for bsz in batch_sizes:
                                                for cl in context_length:
                                                    for (
                                                        use_aot_impl
                                                    ) in use_aot_impl_options:
                                                        for sinks in sinks_options:
                                                            for (
                                                                sliding_window
                                                            ) in sliding_window_options:
                                                                test_config = {
                                                                    "use_torch_flash_ref": use_torch_flash_ref,
                                                                    "compute_type": ct,
                                                                    "quant_q": quant_q,
                                                                    "quant_kv": quant_kv,
                                                                    "trans_v": trans_v_mode,
                                                                    "kv_varlen": kv_varlen_mode,
                                                                    "context_partition_size": context_partition_size,
                                                                    "quant_mode": qm,
                                                                    "block_size": bs,
                                                                    "num_heads": hc,
                                                                    "context_length": cl,
                                                                    "batch_size": bsz,
                                                                    "query_length": ql,
                                                                    "head_size": head_size,
                                                                    "use_aot_impl": use_aot_impl,
                                                                    "sinks": sinks,
                                                                    "sliding_window": sliding_window,
                                                                }

                                                                # Calculate func_name to filter duplicate configs
                                                                (
                                                                    num_query_heads,
                                                                    num_kv_heads,
                                                                ) = hc
                                                                query_group_size = (
                                                                    num_query_heads
                                                                    // num_kv_heads
                                                                )

                                                                # Calculate power of 2 values
                                                                head_size_pow2 = triton.next_power_of_2(
                                                                    head_size
                                                                )

                                                                # Determine quantization modes
                                                                if quant_q:
                                                                    if (
                                                                        qm
                                                                        == "per_tensor"
                                                                    ):
                                                                        query_quant_mode = (
                                                                            0
                                                                        )
                                                                    else:  # per_token
                                                                        query_quant_mode = (
                                                                            1
                                                                        )
                                                                else:
                                                                    query_quant_mode = (
                                                                        -1
                                                                    )

                                                                if quant_kv:
                                                                    if (
                                                                        qm
                                                                        == "per_tensor"
                                                                    ):
                                                                        kv_quant_mode = (
                                                                            0
                                                                        )
                                                                    else:  # per_token
                                                                        kv_quant_mode = (
                                                                            1
                                                                        )
                                                                else:
                                                                    kv_quant_mode = -1

                                                                # Determine fp8_max_value
                                                                if kv_quant_mode >= 0:
                                                                    fp8_max_value = torch.finfo(
                                                                        aiter.dtypes.fp8
                                                                    ).max
                                                                else:
                                                                    fp8_max_value = 1.0

                                                                # Determine is_causal
                                                                is_causal = int(ql > 1)

                                                                # use_sinks for func_name (cdna_version already set at function start)
                                                                use_sinks = 0  # Default: no sinks

                                                                # Calculate func_name
                                                                func_name = get_default_func_name(
                                                                    MD_NAME,
                                                                    (
                                                                        ct,  # compute_type (torch.dtype)
                                                                        ql,  # query_seq_len
                                                                        query_group_size,  # one_query_group_size
                                                                        head_size_pow2,
                                                                        bs,  # kv_block_size
                                                                        context_partition_size,
                                                                        query_quant_mode,
                                                                        kv_quant_mode,
                                                                        fp8_max_value,
                                                                        int(
                                                                            trans_v_mode
                                                                        ),  # value_transposed
                                                                        is_causal,
                                                                        use_sinks,
                                                                        cdna_version,
                                                                    ),
                                                                )
                                                                # Store func_name in test_config for deduplication
                                                                test_config[
                                                                    "func_name"
                                                                ] = func_name
                                                                test_configs.append(
                                                                    test_config
                                                                )
    # Filter out duplicate test configs based on func_name
    total_before_dedup = len(test_configs)
    print(f"\nTotal test cases before deduplication: {total_before_dedup}")

    seen_func_names = set()
    unique_test_configs = []
    for test_config in test_configs:
        func_name = test_config.get("func_name")
        if func_name not in seen_func_names:
            seen_func_names.add(func_name)
            unique_test_configs.append(test_config)

    test_configs = unique_test_configs
    total = len(test_configs)
    print(f"Total test cases after deduplication: {total}")
    print(f"Removed {total_before_dedup - total} duplicate configurations")

    # Prepare arguments for multiprocessing
    test_args = [(config, idx + 1, total) for idx, config in enumerate(test_configs)]

    # Determine number of processes
    if num_processes is None:
        num_processes = min(cpu_count(), total)
    else:
        num_processes = min(cpu_count(), num_processes)
        num_processes = min(total, num_processes)
    print(f"Using {num_processes} parallel processes\n")

    # Run tests in parallel using spawn context to avoid CUDA reinitialization issues
    mp_context = multiprocessing.get_context("spawn")
    with concurrent.futures.ProcessPoolExecutor(
        max_workers=num_processes, mp_context=mp_context
    ) as executor:
        results = list(executor.map(_run_single_test, test_args))

    return pd.DataFrame(results)


def parse_arg_and_run_test():
    """Parse arguments and run tests."""
    print(f"Triton location: {triton}")
    print(f"Triton version: {triton.__version__}")

    parser = create_argument_parser()
    # When running via pytest, use empty args to avoid conflict with pytest's argv
    running_via_pytest = "pytest" in sys.argv[0] or sys.argv[0].endswith("py.test")
    if running_via_pytest:
        args = parser.parse_args([])
    else:
        args = parser.parse_args()
    (
        block_sizes,
        head_configs,
        context_length,
        batch_sizes,
        query_lengths,
        quant_mode,
        trans_v,
        kv_varlen,
        compute_types_quant_q_and_kv,
        use_torch_flash_ref_options,
        use_aot_impl_options,
        context_partition_size_options,
        sinks_options,
        sliding_window_options,
    ) = process_arguments(args)

    run_multi_pa_gluon_test(
        block_sizes,
        head_configs,
        context_length,
        batch_sizes,
        query_lengths,
        quant_mode,
        trans_v,
        kv_varlen,
        compute_types_quant_q_and_kv,
        use_torch_flash_ref_options,
        use_aot_impl_options,
        context_partition_size_options,
        num_processes=args.num_processes,
        sinks_options=sinks_options,
        sliding_window_options=sliding_window_options,
    )
    print("All the so under different configurations have been built successfully!")


def get_so_files_size_and_count():
    """Get the total size and number of so files in aiter build directory."""
    try:
        du_result = subprocess.run(
            ["du", "-sh", BUILD_DIR], capture_output=True, text=True, timeout=100
        )
        if du_result.returncode == 0:
            total_size_of_so_files = du_result.stdout.split()[0]
            print(
                f"The total size of so files in aiter build directory: {total_size_of_so_files}"
            )
    except Exception as e:
        print(
            f"Warning: Could not get the total size of so files in aiter build directory: {e}"
        )
    # Get the number of so files in aiter build directory
    try:
        so_count_result = subprocess.run(
            ["sh", "-c", f"find {BUILD_DIR} -type f -name '*.so' | wc -l"],
            capture_output=True,
            text=True,
            timeout=100,
        )
        if so_count_result.returncode == 0:
            number_of_so_files = so_count_result.stdout.strip()
            print(
                f"The number of so files in aiter build directory: {number_of_so_files}"
            )
    except Exception as e:
        print(
            f"Warning: Could not get the number of so files in aiter build directory: {e}"
        )


def prebuild_normal_accuracy_cases_aot_so():
    """Prebuild normal accuracy cases AOT so files."""
    global BLOCK_SIZE_OPTIONS
    global QUERY_LENGTH_OPTIONS
    global BATCH_SIZE_OPTIONS
    global HEAD_CONFIGURATIONS
    global CONTEXT_LENGTH_OPTIONS
    global COMPUTE_TYPE_OPTIONS
    global QUANT_MODE_OPTIONS
    global HEAD_DIMENSION_OPTIONS
    global TRANS_V_OPTIONS
    global KV_VARLEN_OPTIONS
    global QUANT_Q_AND_KV_OPTIONS
    global USE_TORCH_FLASH_REF_OPTIONS
    global USE_AOT_IMPL_OPTIONS
    global CONTEXT_PARTITION_SIZE_OPTIONS
    global SINKS_OPTIONS
    global SLIDING_WINDOW_OPTIONS
    global COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS

    USE_AOT_IMPL_OPTIONS = [True]
    SINKS_OPTIONS = [False]
    SLIDING_WINDOW_OPTIONS = [0]

    USE_TORCH_FLASH_REF_OPTIONS = [True]
    CONTEXT_PARTITION_SIZE_OPTIONS = [256]
    HEAD_DIMENSION_OPTIONS = [128]
    HEAD_CONFIGURATIONS = [(5, 1), (8, 1), (10, 1), (16, 1)]
    QUERY_LENGTH_OPTIONS = [1, 2, 3, 4]
    COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS = [["fp8", True, True], ["bf16", False, False]]
    QUANT_MODE_OPTIONS = ["per_token", "per_tensor"]
    CONTEXT_LENGTH_OPTIONS = [1027]
    BATCH_SIZE_OPTIONS = [3, 81]
    TRANS_V_OPTIONS = [False]
    KV_VARLEN_OPTIONS = [False, True]
    BLOCK_SIZE_OPTIONS = [16, 64, 1024]
    parse_arg_and_run_test()

    # Test for different head dimensions
    HEAD_DIMENSION_OPTIONS = [64, 192, 256]
    HEAD_CONFIGURATIONS = [(8, 1)]
    QUERY_LENGTH_OPTIONS = [1, 3]
    QUANT_MODE_OPTIONS = ["per_token"]
    BATCH_SIZE_OPTIONS = [81]
    KV_VARLEN_OPTIONS = [True]
    parse_arg_and_run_test()


def prebuild_normal_performance_cases_aot_so():
    """Prebuild normal performance cases AOT so files."""
    global BLOCK_SIZE_OPTIONS
    global QUERY_LENGTH_OPTIONS
    global BATCH_SIZE_OPTIONS
    global HEAD_CONFIGURATIONS
    global CONTEXT_LENGTH_OPTIONS
    global COMPUTE_TYPE_OPTIONS
    global QUANT_MODE_OPTIONS
    global HEAD_DIMENSION_OPTIONS
    global TRANS_V_OPTIONS
    global KV_VARLEN_OPTIONS
    global QUANT_Q_AND_KV_OPTIONS
    global USE_TORCH_FLASH_REF_OPTIONS
    global USE_AOT_IMPL_OPTIONS
    global CONTEXT_PARTITION_SIZE_OPTIONS
    global SINKS_OPTIONS
    global SLIDING_WINDOW_OPTIONS
    global COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS

    USE_AOT_IMPL_OPTIONS = [True]
    SINKS_OPTIONS = [False]
    SLIDING_WINDOW_OPTIONS = [0]

    USE_TORCH_FLASH_REF_OPTIONS = [True]
    CONTEXT_PARTITION_SIZE_OPTIONS = [256]
    HEAD_DIMENSION_OPTIONS = [128]
    CONTEXT_LENGTH_OPTIONS = [2048, 4096, 8192]
    BATCH_SIZE_OPTIONS = [1, 2, 4, 8, 16, 32, 64, 128]
    QUERY_LENGTH_OPTIONS = [1, 2, 3, 4]
    COMPUTE_TYPES_QUANT_Q_AND_KV_OPTIONS = [["fp8", True, True], ["bf16", False, False]]
    QUANT_MODE_OPTIONS = ["per_tensor"]
    TRANS_V_OPTIONS = [False]
    KV_VARLEN_OPTIONS = [False]
    HEAD_CONFIGURATIONS = [(64, 4), (64, 8)]
    BLOCK_SIZE_OPTIONS = [16, 64]
    parse_arg_and_run_test()


if __name__ == "__main__":
    prebuild_normal_accuracy_cases_aot_so()
    prebuild_normal_performance_cases_aot_so()
    get_so_files_size_and_count()



import triton
import triton.language as tl


@triton.jit
def get_topmask_and_fullmask(x):
    tl.static_assert(
        x.dtype.is_int_unsigned(), "floating-point value must be passed as bits"
    )
    tm: tl.constexpr = 1 << (-1 + x.dtype.primitive_bitwidth)
    fm: tl.constexpr = (1 << x.dtype.primitive_bitwidth) - 1
    tm_arr = tl.full(x.shape, tm, dtype=x.dtype)
    fm_arr = tl.full(x.shape, fm, dtype=x.dtype)
    return tm_arr, fm_arr


@triton.jit
def fpval_to_key(x):
    tm, fm = get_topmask_and_fullmask(x)
    return x ^ tl.where((x & tm) != 0, fm, tm)


@triton.jit
def key_to_fpval(x):
    tm, fm = get_topmask_and_fullmask(x)
    return x ^ tl.where((x & tm) == 0, fm, tm)


# stable top-k tie-breaks to value with smaller index
@triton.jit
def indx_to_key(indx, N_EXPTS_PAD: tl.constexpr):
    return N_EXPTS_PAD - indx


@triton.jit
def key_to_indx(indx, N_EXPTS_PAD: tl.constexpr):
    return N_EXPTS_PAD - indx


@triton.jit
def streaming_topk(
    X,
    stride_xm,
    n_expts_tot,
    offs_m,
    mask_m,
    N_EXPTS_PAD: tl.constexpr,
    N_EXPTS_ACT: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    x_nbits: tl.constexpr = X.dtype.element_ty.primitive_bitwidth
    x_utype: tl.constexpr = tl.dtype(f"uint{x_nbits}")
    if x_nbits < 16:
        # this ensures that we leave at least 16 bits for expert index
        # even if the input dtype is smaller than 16 bits:
        y_nbits: tl.constexpr = 32
    else:
        y_nbits: tl.constexpr = x_nbits * 2
    x_ultype: tl.constexpr = tl.dtype(f"uint{y_nbits}")
    x_dtype: tl.constexpr = X.dtype.element_ty

    # subtract 1 from loop iterations because we peel the first (masked) iteration:
    loop_iterations: tl.constexpr = N_EXPTS_PAD // BLOCK_N - 1
    offs_x_n = loop_iterations * BLOCK_N + tl.arange(0, BLOCK_N)
    mask_n = offs_x_n[None, :] < n_expts_tot

    # first iteration:
    X_ptrs = X + offs_m[:, None] * stride_xm + offs_x_n[None, :]
    x = tl.load(X_ptrs, mask=(mask_m & mask_n), other=float("-inf"))
    x = fpval_to_key(x.to(x_utype, bitcast=True))
    x = (x.to(x_ultype) << 16) | indx_to_key(offs_x_n, N_EXPTS_PAD)[None, :]
    acc = tl.topk(x, N_EXPTS_ACT, dim=1)

    # subsequent iterations:
    for _i in (tl.static_range if loop_iterations <= 4 else range)(loop_iterations):
        acc = tl.bitonic_merge(acc)  # ensure sorted ascending for the merge
        X_ptrs -= BLOCK_N
        offs_x_n -= BLOCK_N
        x = tl.load(X_ptrs, mask=mask_m, other=float("-inf"))
        x = fpval_to_key(x.to(x_utype, bitcast=True))
        x = (x.to(x_ultype) << 16) | indx_to_key(offs_x_n, N_EXPTS_PAD)[None, :]
        acc = tl.maximum(acc, tl.topk(x, N_EXPTS_ACT, dim=1))

    # rotate expert index into upper 16 bits:
    # 0000vvvvvvvviiii --> iiii0000vvvvvvvv
    acc = (acc << (y_nbits - 16)) | (acc >> 16)
    # sort in ascending order of expert (descending order of key)
    acc = tl.sort(acc, dim=1, descending=True)
    # iiii0000vvvvvvvv --> 0000iiii:
    y_indices_raw = (acc >> (y_nbits - 16)).to(tl.uint32)
    y_indices = key_to_indx(y_indices_raw, N_EXPTS_PAD)
    # iiii0000vvvvvvvv --> vvvvvvvv:
    y_values_raw = acc.to(x_utype)
    y_values = key_to_fpval(y_values_raw).to(x_dtype, bitcast=True)

    return y_values, y_indices


@triton.jit
def _topk(
    X,
    stride_xm,  # inputs
    Yv,
    Yi,
    stride_ym,  # topk values/indices
    USE_PROVIDED_INDX: tl.constexpr,
    Bits,
    stride_rm,
    stride_rn,  # bitmatrix
    n_rows,
    n_expts_tot,  # shape
    S,
    BLOCK_S: tl.constexpr,
    s_blocks,  # thing to memset
    SP,
    BLOCK_SP: tl.constexpr,
    sp_blocks,
    sp_size,
    APPLY_SOFTMAX: tl.constexpr,  # constant
    BLOCK_M: tl.constexpr,
    N_EXPTS_PAD: tl.constexpr,
    N_EXPTS_ACT: tl.constexpr,
    BLOCK_N: tl.constexpr,
):

    pid = tl.program_id(0)
    if isinstance(n_rows, tl.tensor) and n_rows.dtype.is_ptr():
        n_rows = tl.load(n_rows)

    if pid < s_blocks:
        tl.store(
            S + BLOCK_S * pid + tl.arange(0, BLOCK_S), tl.zeros([BLOCK_S], tl.int32)
        )
    elif pid < s_blocks + sp_blocks:
        offs = BLOCK_SP * (pid - s_blocks) + tl.arange(0, BLOCK_SP)
        tl.store(SP + offs, tl.zeros([BLOCK_SP], tl.int32), mask=offs < sp_size)

    if pid * BLOCK_M >= n_rows:
        # early exit:
        return

    tl.static_assert(BLOCK_N % 32 == 0)
    tl.static_assert(N_EXPTS_PAD % BLOCK_N == 0)
    x_dtype: tl.constexpr = X.dtype.element_ty

    # load logits
    offs_m = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_y_n = tl.arange(0, N_EXPTS_ACT)
    mask_m = offs_m[:, None] < n_rows
    if USE_PROVIDED_INDX:
        Yi_ptrs = Yi + offs_m[:, None] * stride_ym + offs_y_n[None, :]
        y_indices = tl.load(Yi_ptrs, mask=mask_m)
        Xv_ptrs = X + offs_m[:, None] * stride_xm + y_indices
        y_values = tl.load(Xv_ptrs, mask=mask_m)
    else:
        y_values, y_indices = streaming_topk(
            X,
            stride_xm,
            n_expts_tot,
            offs_m,
            mask_m,  #
            N_EXPTS_PAD,
            N_EXPTS_ACT,
            BLOCK_N,
        )

    # normalize selected values
    if APPLY_SOFTMAX:
        y_values = tl.softmax(y_values.to(tl.float32), dim=1, keep_dims=True).to(
            x_dtype
        )

    # write back
    Yv_ptrs = Yv + offs_m[:, None] * stride_ym + offs_y_n[None, :]
    tl.store(Yv_ptrs, y_values, mask=mask_m)
    if not USE_PROVIDED_INDX:
        Yi_ptrs = Yi + offs_m[:, None] * stride_ym + offs_y_n[None, :]
        tl.store(Yi_ptrs, y_indices, mask=mask_m)

    # pack into bitmatrix
    y_div = y_indices // 32
    y_rem = y_indices % 32
    loop_iterations = N_EXPTS_PAD // BLOCK_N
    for i in range(loop_iterations):
        offs_r_n = tl.arange(0, BLOCK_N // 32) + i * (BLOCK_N // 32)
        y2 = tl.where(
            y_div[:, :, None] == offs_r_n[None, None, :], (1 << y_rem)[:, :, None], 0
        )
        r = tl.reduce_or(y2, axis=1)
        BitsPtrs = Bits + offs_m[:, None] * stride_rm + offs_r_n[None, :] * stride_rn
        tl.store(BitsPtrs, r, mask=mask_m)
